# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for Debusine CLI "collection relation" commands."""

import contextlib
import io
from typing import Any, override
from unittest import mock

import yaml

from debusine.artifacts.models import CollectionCategory
from debusine.client.commands import collection_relations
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.exceptions import DebusineError
from debusine.client.models import (
    CollectionIdentifierRequest,
    CollectionIdentifierResponse,
    CollectionRelationResponse,
    CollectionRelationSetResponse,
    CollectionRelationType,
)
from debusine.utils.input import YamlEditor


class ListTests(BaseCliTests):
    """Tests for listing collection relations."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.collection_relation = CollectionRelationResponse(
            source_collection=CollectionIdentifierResponse(
                name="personal", category=CollectionCategory.SUITE
            ),
            target_collection=CollectionIdentifierResponse(
                name="trixie", category=CollectionCategory.SUITE
            ),
            type=CollectionRelationType.SUITE_REQUIRES,
            position=1,
        )

    def test_list(self) -> None:
        command = self.create_command(
            [
                "collection",
                "relation",
                "list",
                "--from",
                "personal@debian:suite",
                "--to",
                "trixie@debian:suite",
                "--type",
                CollectionRelationType.SUITE_REQUIRES,
                "--workspace",
                "workspace",
            ]
        )

        assert isinstance(command, collection_relations.List)

        with (
            mock.patch.object(
                command.debusine,
                "collection_relation_list",
                return_value=[self.collection_relation],
            ) as collection_relation_list_mock,
            mock.patch.object(command, "list") as list_,
        ):
            command.run()

        collection_relation_list_mock.assert_called_once_with(
            "workspace",
            source="personal@debian:suite",
            target="trixie@debian:suite",
            relation_type=CollectionRelationType.SUITE_REQUIRES,
        )

        list_.assert_called_once_with([self.collection_relation])

    def test_list_rich(self) -> None:
        """Test list_rich() method: mocks rich, assert is called as expected."""
        command = self.create_command(
            [
                "collection",
                "relation",
                "list",
                "--from",
                "personal@debian:suite",
                "--to",
                "trixie@debian:suite",
                "--type",
                CollectionRelationType.SUITE_REQUIRES,
                "--workspace",
                "workspace",
            ]
        )

        assert isinstance(command, collection_relations.List)

        with (
            mock.patch(
                "debusine.client.commands.collection_relations.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.collection_relations.rich.print"
            ) as rich_print,
        ):
            command._list_rich([self.collection_relation])

        rich_print.assert_called_once()

        table.add_row.assert_called_once_with(
            "personal@debian:suite",
            "trixie@debian:suite",
            CollectionRelationType.SUITE_REQUIRES,
            "1",
        )

    def test_invalid_format_collection_identifier(self) -> None:
        stderr = io.StringIO()

        with (
            self.assertRaisesRegex(SystemExit, "2"),
            contextlib.redirect_stderr(stderr),
        ):
            self.create_command(
                [
                    "collection",
                    "relation",
                    "list",
                    "--from",
                    "invalid",
                    "--to",
                    "trixie@debian:suite",
                    "--type",
                    CollectionRelationType.SUITE_REQUIRES,
                    "--workspace",
                    "workspace",
                ],
            )

        self.assertIn(
            "Invalid collection identifier: 'invalid'. "
            "Expected format is 'name@category'",
            stderr.getvalue(),
        )

    def test_invalid_collection_identifier(self) -> None:
        stderr = io.StringIO()

        with (
            self.assertRaisesRegex(SystemExit, "2"),
            contextlib.redirect_stderr(stderr),
        ):
            self.create_command(
                [
                    "collection",
                    "relation",
                    "list",
                    "--from",
                    "personal@invalid",
                    "--to",
                    "trixie@debian:suite",
                    "--type",
                    CollectionRelationType.SUITE_REQUIRES,
                    "--workspace",
                    "workspace",
                ],
            )

        self.assertIn(
            "Invalid collection identifier: 'personal@invalid'. "
            "Category must be one of: debian:archive, debian:environments, "
            "debian:package-build-logs, debian:qa-results, debian:suite, "
            "debusine:task-configuration, debusine:task-history, "
            "debusine:workflow-internal",
            stderr.getvalue(),
        )

    def test_invalid_collection_identifier_debusine_test(self) -> None:
        stderr = io.StringIO()

        with (
            self.assertRaisesRegex(SystemExit, "2"),
            contextlib.redirect_stderr(stderr),
        ):
            self.create_command(
                [
                    "collection",
                    "relation",
                    "list",
                    "--from",
                    "test@debusine:test",
                    "--to",
                    "trixie@debian:suite",
                    "--type",
                    CollectionRelationType.SUITE_REQUIRES,
                    "--workspace",
                    "workspace",
                ],
            )

        self.assertIn(
            "Invalid collection identifier: 'test@debusine:test'. "
            "Category must be one of: debian:archive, debian:environments, "
            "debian:package-build-logs, debian:qa-results, debian:suite, "
            "debusine:task-configuration, debusine:task-history, "
            "debusine:workflow-internal",
            stderr.getvalue(),
        )

    def test_invalid_collection_identifier_wrong_format(self) -> None:
        stderr = io.StringIO()

        with (
            self.assertRaisesRegex(SystemExit, "2"),
            contextlib.redirect_stderr(stderr),
        ):
            self.create_command(
                [
                    "collection",
                    "relation",
                    "list",
                    "--from",
                    "a@",
                    "--to",
                    "trixie@debian:suite",
                    "--type",
                    CollectionRelationType.SUITE_REQUIRES,
                    "--workspace",
                    "workspace",
                ],
            )

        self.assertIn(
            "Invalid collection identifier: 'a@'. "
            "Expected format is 'name@category'",
            stderr.getvalue(),
        )


class EditTests(BaseCliTests):
    """Tests for editing collection relations."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.source = "personal@debian:suite"
        self.relation_type = CollectionRelationType.SUITE_REQUIRES
        self.current_targets = [
            CollectionIdentifierResponse(
                name="trixie", category=CollectionCategory.SUITE
            ),
            CollectionIdentifierResponse(
                name="experimental", category=CollectionCategory.SUITE
            ),
        ]
        self.etag = '"87d154802e0803b953e5fee3e"'

    def create_edit_command(
        self, *args: str, relation_type: None | str = None
    ) -> collection_relations.Edit:
        if relation_type is None:
            relation_type = self.relation_type

        command = self.create_command(
            [
                "collection",
                "relation",
                "edit",
                self.source,
                relation_type,
                "--workspace",
                "workspace",
                *args,
            ]
        )
        assert isinstance(command, collection_relations.Edit)
        return command

    def current_relations(self) -> list[CollectionRelationResponse]:
        return [
            CollectionRelationResponse(
                source_collection=CollectionIdentifierResponse(
                    name="personal", category=CollectionCategory.SUITE
                ),
                target_collection=target,
                type=self.relation_type,
                position=index + 1,
            )
            for index, target in enumerate(self.current_targets)
        ]

    def capture_run_mock_api(
        self,
        *args: str,
        assert_system_exit_code: int | None = None,
        collection_relation_set_side_effect: Any | None = None,
    ) -> tuple[mock.MagicMock, mock.MagicMock, str, str]:
        command = self.create_edit_command(*args)

        if collection_relation_set_side_effect is not None:
            set_side_effect = [collection_relation_set_side_effect]
        else:
            set_side_effect = [
                CollectionRelationSetResponse(targets=self.current_targets)
            ]

        with (
            mock.patch.object(
                command.debusine,
                "collection_relation_list_for_edit",
                return_value=(self.current_relations(), self.etag),
            ) as collection_relation_list_mock,
            mock.patch.object(
                command.debusine,
                "collection_relation_set",
                side_effect=set_side_effect,
            ) as collection_relation_set_mock,
        ):
            stderr, stdout = self.capture_output(
                command.run, assert_system_exit_code=assert_system_exit_code
            )

        return (
            collection_relation_list_mock,
            collection_relation_set_mock,
            stderr,
            stdout,
        )

    def test_edit_incompatible_other_options(self) -> None:
        command = self.create_edit_command(
            "--append",
            "bookworm@debian:suite",
            "--edit",
            relation_type="suite-requires",
        )

        stderr, stdout = self.capture_output(
            command.run, assert_system_exit_code=1
        )

        self.assertEqual(
            stderr,
            "--edit cannot be combined with -"
            "-append, --prepend, --remove, or --set\n",
        )
        self.assertEqual(stdout, "")

    def test_no_change(self) -> None:
        list_, set_, stderr, stdout = self.capture_run_mock_api()

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        list_.assert_called_once_with(
            "workspace",
            source=self.source,
            relation_type=self.relation_type,
        )
        set_.assert_not_called()

    def test_append(self) -> None:
        list_, set_, stderr, stdout = self.capture_run_mock_api(
            "--append", "bookworm@debian:suite"
        )

        self.assertEqual(stderr, "")
        list_.assert_called_once()
        set_.assert_called_once_with(
            "workspace",
            source=self.source,
            relation_type=self.relation_type,
            targets=[
                CollectionIdentifierRequest(
                    name="trixie", category=CollectionCategory.SUITE
                ),
                CollectionIdentifierRequest(
                    name="experimental", category=CollectionCategory.SUITE
                ),
                CollectionIdentifierRequest(
                    name="bookworm", category=CollectionCategory.SUITE
                ),
            ],
            if_match=self.etag,
        )
        self.assertEqual(
            yaml.safe_load(stdout),
            {
                "targets": [
                    target.model_dump() for target in self.current_targets
                ]
            },
        )
        self.assertEqual(stderr, "")

    def test_prepend_remove_set(self) -> None:
        list_, set_, stderr, stdout = self.capture_run_mock_api(
            "--set",
            "bookworm@debian:suite",
            "trixie@debian:suite",
            "--prepend",
            "sid@debian:suite",
            "--remove",
            "trixie@debian:suite",
        )

        stdout_parsed = yaml.safe_load(stdout)

        self.assertEqual(
            stdout_parsed,
            {
                "targets": [
                    target.model_dump() for target in self.current_targets
                ]
            },
        )

        self.assertEqual(stderr, "")
        list_.assert_called_once()
        set_.assert_called_once_with(
            "workspace",
            source=self.source,
            relation_type=self.relation_type,
            targets=[
                CollectionIdentifierRequest(
                    name="sid", category=CollectionCategory.SUITE
                ),
                CollectionIdentifierRequest(
                    name="bookworm", category=CollectionCategory.SUITE
                ),
            ],
            if_match=self.etag,
        )

    def test_edit_no_changes(self) -> None:
        with mock.patch(
            "debusine.utils.input.YamlEditor.edit_apply_loop",
            autospec=True,
            return_value=None,
        ) as edit_apply_loop_mock:
            list_, set_, stderr, stdout = self.capture_run_mock_api("--edit")

        edit_apply_loop_mock.assert_called_once()

        # Listing called once
        list_.assert_called_once_with(
            "workspace",
            source="personal@debian:suite",
            relation_type=CollectionRelationType.SUITE_REQUIRES,
        )

        # No set called
        set_.assert_not_called()

        self.assertEqual(stderr, "")
        self.assertEqual(
            stdout, "- trixie@debian:suite\n- experimental@debian:suite\n"
        )

    def test_edit(self) -> None:
        edited_targets = ["another@debian:suite"]

        def fake_edit(editor: YamlEditor[list[Any]]) -> bool:
            editor.value = edited_targets
            return True

        with mock.patch(
            "debusine.utils.input.YamlEditor.edit",
            autospec=True,
            side_effect=fake_edit,
        ) as edit_mock:
            list_, set_, stderr, stdout = self.capture_run_mock_api("--edit")

        edit_mock.assert_called_once()

        list_.assert_called_once_with(
            "workspace",
            source="personal@debian:suite",
            relation_type=CollectionRelationType.SUITE_REQUIRES,
        )

        set_.assert_called_once_with(
            "workspace",
            source="personal@debian:suite",
            relation_type=CollectionRelationType.SUITE_REQUIRES,
            targets=[
                CollectionIdentifierRequest(
                    name="another", category=CollectionCategory.SUITE
                )
            ],
            if_match=self.etag,
        )

        self.assertEqual(
            yaml.safe_load(stdout),
            {
                "targets": [
                    target.model_dump() for target in self.current_targets
                ]
            },
        )
        self.assertEqual(stderr, "")

    def test_api_fail(self) -> None:
        _list, _set, stderr, stdout = self.capture_run_mock_api(
            "--prepend",
            "sid@debian:suite",
            "--remove",
            "trixie@debian:suite",
            "--set",
            "bookworm@debian:suite",
            "trixie@debian:suite",
            assert_system_exit_code=3,
            collection_relation_set_side_effect=DebusineError(title="Error DB"),
        )
        stderr_parsed = yaml.safe_load(stderr)

        self.assertEqual(stderr_parsed["error"]["title"], "Error DB")
        self.assertEqual(stdout, "")



# Generated by Django 4.2.28 on 2026-04-30 09:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0053_workrequesttoken'),
    ]

    operations = [
        migrations.AlterField(
            model_name='scoperole',
            name='role',
            field=models.CharField(
                choices=[('owner', 'Owner'), ('observer', 'Observer')]
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""DB constraints."""

from typing import Any, override

from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.backends.ddl_references import Statement, Table
from django.db.models import Model, Q
from django.db.models.constraints import BaseConstraint
from django.db.models.sql import Query


class JsonDataUniqueConstraint(BaseConstraint):
    """UniqueConstraint for JsonField data."""

    def __init__(
        self,
        *,
        fields: list[str] | tuple[str, ...],
        name: str,
        condition: Q | None = None,
        nulls_distinct: bool | None = None,
    ) -> None:
        """
        Create a UniqueConstraint operating on a JsonField.

        :param fields: list of keys in the JsonField that must be unique
        :param condition: condition for the constraint to be enforced
           (e.g. if the constraint is only enforced for some "category")
        :param name: constraint name
        :param nulls_distinct: False: nulls are not distinct.
           If nulls_distinct=False requires non-JSON fields to be null=False.
        """
        self.name = name
        self._condition = condition
        self._fields = tuple(fields)
        self._nulls_distinct = nulls_distinct
        super().__init__(name)

    def _get_condition_sql(
        self, model: type[Model], schema_editor: BaseDatabaseSchemaEditor
    ) -> str | None:
        # Copy-paste of UniqueConstraint._get_condition_sql
        if self._condition is None:
            return None
        query = Query(model=model, alias_cols=False)
        where = query.build_where(self._condition)
        compiler = query.get_compiler(connection=schema_editor.connection)
        sql, params = where.as_sql(compiler, schema_editor.connection)
        return sql % tuple(schema_editor.quote_value(p) for p in params)

    @override
    def constraint_sql(
        self,
        model: type[Model] | None,
        schema_editor: BaseDatabaseSchemaEditor | None,
    ) -> str:
        """Not currently used."""
        raise NotImplementedError()

    @override
    def create_sql(
        self,
        model: type[Model] | None,
        schema_editor: BaseDatabaseSchemaEditor | None,
    ) -> str:
        """Return the statement to create the index."""
        assert model
        assert schema_editor

        if schema_editor.connection.vendor != "postgresql":
            raise NotImplementedError(
                f"{self.__class__.__name__} only supports PostgreSQL"
            )

        columns = []

        for field in self._fields:
            if "->>" in field:
                columns.append(f"({field})")
            else:
                columns.append(
                    model._meta.get_field(field).column  # type: ignore[arg-type,union-attr]
                )

        condition = self._get_condition_sql(model, schema_editor)

        if self._nulls_distinct is True:
            nulls_distinct = " NULLS DISTINCT"
        elif self._nulls_distinct is False:
            nulls_distinct = " NULLS NOT DISTINCT"
        else:
            nulls_distinct = ""

        if condition:
            where = f" WHERE {condition}"
        else:
            where = ""

        table_quoted = Table(
            model._meta.db_table, schema_editor.connection.ops.quote_name
        )

        sql = (
            "CREATE UNIQUE INDEX %(name)s ON %(table)s (%(columns)s)"
            "%(nulls_distinct)s%(where)s"
        )

        return Statement(
            sql,
            name=self.name,
            table=table_quoted,
            columns=", ".join(columns),
            nulls_distinct=nulls_distinct,
            where=where,
        )  # type: ignore[return-value]

    @override
    def remove_sql(
        self,
        model: type[Model] | None,  # noqa: ARG002, U100
        schema_editor: BaseDatabaseSchemaEditor | None,  # noqa: ARG002, U100
    ) -> str:
        """Remove the index."""
        sql = "DROP INDEX %(name)s"
        return Statement(sql, name=self.name)  # type: ignore[return-value]

    @override
    def deconstruct(self) -> tuple[str, tuple[Any], dict[str, Any]]:
        """
        Deconstruct the index.

        Return what is serialized in the migration file.
        """
        path, args, kwargs = super().deconstruct()

        path = path.replace(
            "debusine.db.models.constraints", "debusine.db.models"
        )

        kwargs["fields"] = self._fields
        if self._condition:
            kwargs["condition"] = self._condition
        if self._nulls_distinct is not None:
            kwargs["nulls_distinct"] = self._nulls_distinct

        return path, tuple(args), kwargs

    @override
    def __eq__(self, other: Any) -> bool:
        """Return True if other is equal to self."""
        if isinstance(other, JsonDataUniqueConstraint):
            return (
                self._fields == other._fields
                and self._condition == other._condition
                and self._nulls_distinct is other._nulls_distinct
                and self.name == other.name
            )

        return super().__eq__(other)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""
Base Workflow infrastructure.

See :ref:`explanation-workflows` for a high level explanation of concepts used
here.
"""

import logging
from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from functools import partial
from typing import Any, Literal, cast, override

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    TaskTypes,
    WorkRequestResults,
)
from debusine.client.models import LookupChildType, RuntimeParameter
from debusine.db.models import (
    Collection,
    CollectionItem,
    WorkRequest,
    Workspace,
)
from debusine.db.models.work_requests import (
    InternalTaskError,
    StatusChangeError,
)
from debusine.db.tasks import DBTask
from debusine.server.workflows.models import (
    BaseWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.tasks import TaskConfigError, inputs
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    ActionUpdateCollectionWithData,
    BaseDynamicTaskData,
    BaseTaskData,
    InputArtifact,
    LookupMultiple,
    LookupSingle,
    OutputData,
    OutputDataError,
)

logger = logging.getLogger(__name__)


class WorkflowValidationError(Exception):
    """Raised if a workflow fails to validate its inputs."""


class WorkflowRunError(Exception):
    """Running a workflow orchestrator or callback failed."""

    def __init__(
        self, work_request: WorkRequest, message: str, code: str
    ) -> None:
        """Construct the exception."""
        self.work_request = work_request
        self.message = message
        self.code = code


class Workflow[
    WD: BaseWorkflowData,
    DTD: BaseDynamicTaskData = BaseDynamicTaskData,
](DBTask[WD, DTD], metaclass=ABCMeta):
    """
    Base class for workflow orchestrators.

    This is the base API for running :py:class:`WorkflowInstance` logic.
    """

    TASK_TYPE = TaskTypes.WORKFLOW

    #: Internal work request when instantiated from a workflow callback
    work_request_callback: WorkRequest | None

    @override
    def __init__(self, work_request: WorkRequest) -> None:
        super().__init__(work_request)
        self.work_request_callback = None

    @classmethod
    def from_name(cls, name: str) -> type['Workflow[Any, Any]']:
        """Instantiate a workflow by name."""
        res = super().class_from_name(TaskTypes.WORKFLOW, name)
        return cast(type[Workflow[Any, Any]], res)

    @classmethod
    def validate_static_parameters(cls, data: dict[str, Any]) -> None:
        """Validate WorkflowTemplate static_parameters."""
        # By default nothing is enforced

    @classmethod
    def validate_runtime_parameters(
        cls, data: dict[str, Any] | Literal[RuntimeParameter.ANY]
    ) -> None:
        """Validate WorkflowTemplate runtime_parameters."""
        if data == RuntimeParameter.ANY:
            return
        assert isinstance(data, dict)
        for key, value in data.items():
            if key not in cls.task_data_type.model_fields:
                raise WorkflowValidationError(
                    f"{key} is not a task_data parameter for {cls.__name__}"
                )
            if value == RuntimeParameter.ANY:
                continue
            if isinstance(value, str):
                raise WorkflowValidationError(
                    f"{key} is set to a string ({value!r}), not a list of "
                    f"valid values."
                )
            if isinstance(value, list) and all(
                isinstance(x, str) for x in value
            ):
                continue
            raise WorkflowValidationError(
                f"{key} is set to {value!r}, not a list of valid values."
            )

    def validate_input(self) -> None:
        """
        Thorough validation of input data.

        This is run only once at workflow instantiation time, and can do
        slower things like database lookups to validate artifact or collection
        types.
        """
        # Do nothing by default

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()

        # Compute dynamic task to extract source package
        # TODO: these values can be computed in a separate methods if the rest
        # TODO: of compute_dynamic_data becomes too expensive to run twice
        dynamic_data = self.compute_dynamic_data()

        if package_name := dynamic_data.get_source_package_name():
            provided.add(
                ttags.ProvenanceProvided.SYSTEM,
                {ttags.SOURCE_PACKAGE_PREFIX + package_name},
            )

        return provided, required

    def ensure_dynamic_data(self) -> None:
        """Ensure that this workflow's dynamic task data has been computed."""
        if self.work_request.dynamic_task_data is None:
            # TODO: this could be moved to when the workflow is marked running,
            # to match the invariant for other task types that once a task is
            # running then their dynamic data is populated
            self.resolve_inputs(inputs.Stage.RUNNING)
            dynamic_task_data = self.compute_dynamic_data()
            self.work_request.dynamic_task_data = dynamic_task_data.model_dump(
                mode="json", exclude_unset=True
            )
            self.work_request.save()
            self.dynamic_data = self.dynamic_task_data_type(
                **self.work_request.dynamic_task_data
            )

    @override
    def get_label(self) -> str:
        """Return a short human-readable label for the task."""
        name: str

        if self.work_request_callback is not None:
            workflow_data = self.work_request_callback.workflow_data
            name = (
                workflow_data.display_name
                or workflow_data.step
                or self.work_request.workflow_data.workflow_template_name
                or self.name
            )
        else:
            name = (
                self.work_request.workflow_data.workflow_template_name
                or self.name
            )

        if self.dynamic_data is None:
            return name

        if self.dynamic_data.parameter_summary is None:
            return name

        return f"{name}: {self.dynamic_data.parameter_summary}"

    @override
    def get_description(self) -> str:
        if self.work_request_callback is not None:
            if step := self.work_request_callback.workflow_data.step:
                try:
                    cb = self.get_callback(step)
                except NotImplementedError:
                    pass
                else:
                    if cb.__doc__:
                        return cb.__doc__
        return super().get_description()

    @abstractmethod
    def populate(self) -> None:
        """
        Create the initial WorkRequest structure.

        This is called once, when the workflow first becomes runnable.
        :py:meth:`validate_input` will already have been called.

        This method is required to be idempotent: calling it multiple times
        with the same argument MUST result in the same
        :py:class:`WorkRequest` structure as calling it once.
        """

    def get_callback(self, step: str) -> Callable[[], bool]:
        """
        Get the callback implementation for a given workflow step name.

        By default, subclasses may provide ``callback_{step.replace('-',
        '_')}`` methods to implement workflow callbacks with the
        corresponding ``step`` set in their workflow data.  Alternatively,
        they may override this method to provide a more complex mapping from
        work request to callback implementation.
        """
        callback_method = getattr(
            self, f"callback_{step.replace('-', '_')}", None
        )
        if callback_method is None:
            raise NotImplementedError(
                f"Unhandled workflow callback step: {step}"
            )
        assert callable(callback_method)
        # https://github.com/python/mypy/issues/20748
        return callback_method  # type: ignore[no-any-return]

    def callback(self, work_request: "WorkRequest") -> bool:
        """
        Perform an orchestration step.

        Called with the workflow callback work request (note that this is
        not the same as ``self.work_request``) when the workflow node
        becomes ready to execute.

        Called with a :py:class:`WorkRequest` of type internal/workflow to
        perform an orchestration step triggered by a workflow callback.

        This method is required to be idempotent: calling it multiple times
        with the same argument MUST result in the same
        :py:class:`WorkRequest` structure as calling it once.

        Subclasses may provide ``callback_{step.replace('-', '_')}`` methods
        to implement workflow callbacks with the corresponding ``step`` set
        in their workflow data, or override ``get_callback``.

        :return: True if the callback succeeded, otherwise False.
        """
        step = work_request.workflow_data.step
        if step is None:
            raise NotImplementedError(
                "Workflow callback called without setting `step`"
            )
        succeeded = self.get_callback(step)()
        assert isinstance(succeeded, bool)
        return succeeded

    def orchestrate_child(self, work_request: WorkRequest) -> None:
        """
        Orchestrate a child work request in whatever way is appropriate.

        Silently skips non-pending/running work requests.
        """
        if (
            work_request.status == WorkRequest.Statuses.BLOCKED
            and work_request.can_be_automatically_unblocked()
        ):
            work_request.mark_pending()

        if work_request.status in {
            WorkRequest.Statuses.PENDING,
            WorkRequest.Statuses.RUNNING,
        }:
            # This may result in orchestrator errors, but if so the errors
            # will be logged.  Workflows should not immediately propagate
            # the error; they will eventually be marked as COMPLETED/ERROR
            # once all their children have completed or aborted.
            orchestrate_workflow(work_request)

    @staticmethod
    def provides_artifact(
        work_request: WorkRequest,
        category: ArtifactCategory,
        name: str,
        *,
        data: dict[str, Any] | None = None,
        artifact_filters: dict[str, Any] | None = None,  # noqa: U100
    ) -> None:
        """
        Indicate work_request will provide an artifact.

        :param work_request: work request that will provide the artifact
        :param category: category of the artifact that will be provided
        :param name: name of this item in the workflow’s internal collection
        :param data: add it to the data dictionary for the event reaction
        :param artifact_filters: for the
          :ref:`action-update-collection-with-artifacts` action, to allow
          workflows to add filtering
        :raise LookupError: if a key in "data" starts with ``promise_``

        Create an event reaction for ``on_creation`` adding a promise: this
        work request will create an artifact.

        Create an event reaction for ``on_success`` to update the collection
        with the relevant artifact.
        """
        if "/" in name:
            # This wouldn't work well, because "/" is used to separate
            # lookup string segments.
            raise ValueError('Collection item name may not contain "/".')

        if data is not None:
            for key in data:
                if key.startswith("promise_"):
                    raise ValueError(
                        f'Field name "{key}" starting with '
                        f'"promise_" is not allowed.'
                    )

        # work_request is part of a workflow
        assert work_request.parent is not None

        try:
            work_request.lookup_single(
                f"internal@collections/name:{name}",
                expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
            )
        except KeyError:
            work_request.process_update_collection_with_data(
                [
                    ActionUpdateCollectionWithData(
                        collection="internal@collections",
                        name_template=name,
                        category=BareDataCategory.PROMISE,
                        data={
                            "promise_work_request_id": work_request.id,
                            "promise_workflow_id": work_request.parent.id,
                            "promise_category": category,
                            **(data or {}),
                        },
                    )
                ]
            )

        artifact_filters = artifact_filters or {}
        work_request.add_event_reaction(
            "on_success",
            ActionUpdateCollectionWithArtifacts(
                collection="internal@collections",
                name_template=name,
                variables=data,
                artifact_filters={**artifact_filters, "category": category},
            ),
        )

    @staticmethod
    def requires_artifact(
        work_request: WorkRequest, lookup: LookupSingle | LookupMultiple
    ) -> None:
        """
        Indicate that work_request requires input (lookup).

        :param work_request: for each lookup result call
          ``work_request.add_dependency(promise["promise_work_request_id"])``
        :param lookup: resolve the lookup and iterate over the results
          (for PROMISES only)
        """
        if isinstance(lookup, LookupMultiple):
            results = work_request.lookup_multiple(
                lookup, expect_type=LookupChildType.ARTIFACT_OR_PROMISE
            )
        else:
            results = [
                work_request.lookup_single(
                    lookup, expect_type=LookupChildType.ARTIFACT_OR_PROMISE
                )
            ]

        for result in results:
            collection_item = result.collection_item

            if (
                collection_item is not None
                and collection_item.child_type == CollectionItem.Types.BARE
                and collection_item.category == BareDataCategory.PROMISE
            ):
                work_request_dependency = WorkRequest.objects.get(
                    id=collection_item.data["promise_work_request_id"]
                )
                work_request.add_dependency(work_request_dependency)

    def work_request_ensure_child_worker(
        self,
        *,
        task_name: str,
        task_data: BaseTaskData | None = None,
        workflow_data: WorkRequestWorkflowData,
        relative_priority: int = 0,
    ) -> tuple[WorkRequest, bool]:
        """
        Create a child WORKER work request if one does not already exist.

        This assumes that ``workflow_data.step`` is unique among work
        request children with this task type and name, and that it is stable
        even if the workflow is orchestrated multiple times.

        :param task_name: the task name for the child work request.
        :param task_data: the task data for the child work request.
        :param workflow_data: the workflow data for the child work request.
        :param relative_priority: if creating a new work request, set the
          base priority of the child to the effective priority of this
          workflow plus this relative priority.
        :return: tuple of ``(object, created)``, where ``object`` is the new
          or existing :py:class:`WorkRequest` and ``created`` is a boolean
          specifying whether a new object was created.
        """
        try:
            # Index: db_workrequest_wf_child_idx.
            return (
                self.work_request.children.unsuperseded().get(
                    task_type=TaskTypes.WORKER,
                    task_name=task_name,
                    workflow_data_json__step=workflow_data.step,
                ),
                False,
            )
        except WorkRequest.DoesNotExist:
            return (
                self.work_request.create_child_worker(
                    task_name=task_name,
                    task_data=task_data,
                    workflow_data=workflow_data,
                    relative_priority=relative_priority,
                ),
                True,
            )

    def work_request_ensure_child_server(
        self,
        *,
        task_name: str,
        task_data: BaseTaskData | None = None,
        workflow_data: WorkRequestWorkflowData,
    ) -> tuple[WorkRequest, bool]:
        """
        Create a child SERVER work request if one does not already exist.

        This assumes that ``workflow_data.step`` is unique among work
        request children with this task type and name, and that it is stable
        even if the workflow is orchestrated multiple times.

        :param task_name: the task name for the child work request.
        :param task_data: the task data for the child work request.
        :param workflow_data: the workflow data for the child work request.
        :return: tuple of ``(object, created)``, where ``object`` is the new
          or existing :py:class:`WorkRequest` and ``created`` is a boolean
          specifying whether a new object was created.
        """
        try:
            # Index: db_workrequest_wf_child_idx.
            return (
                self.work_request.children.unsuperseded().get(
                    task_type=TaskTypes.SERVER,
                    task_name=task_name,
                    workflow_data_json__step=workflow_data.step,
                ),
                False,
            )
        except WorkRequest.DoesNotExist:
            return (
                self.work_request.create_child_server(
                    task_name=task_name,
                    task_data=task_data,
                    workflow_data=workflow_data,
                ),
                True,
            )

    def work_request_ensure_child_internal(
        self,
        *,
        task_name: str,
        workflow_data: WorkRequestWorkflowData,
        unblock_strategy: WorkRequest.UnblockStrategy = (
            WorkRequest.UnblockStrategy.DEPS
        ),
        relative_priority: int = 0,
    ) -> tuple[WorkRequest, bool]:
        """
        Create a child INTERNAL work request if one does not already exist.

        This assumes that ``workflow_data.step`` is unique among work
        request children with this task type and name, and that it is stable
        even if the workflow is orchestrated multiple times.

        :param task_name: the task name for the child work request.
        :param task_data: the task data for the child work request.
        :param workflow_data: the workflow data for the child work request.
        :param unblock_strategy: the unblock strategy for the child work
          request.
        :param relative_priority: if creating a new work request, set the
          base priority of the child to the effective priority of this
          workflow plus this relative priority.
        :return: tuple of ``(object, created)``, where ``object`` is the new
          or existing :py:class:`WorkRequest` and ``created`` is a boolean
          specifying whether a new object was created.
        """
        try:
            # Index: db_workrequest_wf_child_idx.
            return (
                self.work_request.children.unsuperseded().get(
                    task_type=TaskTypes.INTERNAL,
                    task_name=task_name,
                    workflow_data_json__step=workflow_data.step,
                ),
                False,
            )
        except WorkRequest.DoesNotExist:
            return (
                self.work_request.create_child_internal(
                    task_name=task_name,
                    workflow_data=workflow_data,
                    unblock_strategy=unblock_strategy,
                    relative_priority=relative_priority,
                ),
                True,
            )

    def work_request_ensure_child_workflow(
        self,
        *,
        task_name: str,
        task_data: BaseTaskData | None = None,
        workflow_data: WorkRequestWorkflowData,
        relative_priority: int = 0,
    ) -> tuple[WorkRequest, bool]:
        """
        Create a child WORKFLOW work request if one does not already exist.

        This assumes that ``workflow_data.step`` is unique among work
        request children with this task type and name, and that it is stable
        even if the workflow is orchestrated multiple times.

        :param task_name: the task name for the child work request.
        :param task_data: the task data for the child work request.
        :param workflow_data: the workflow data for the child work request.
        :param relative_priority: if creating a new work request, set the
          base priority of the child to the effective priority of this
          workflow plus this relative priority.
        :return: tuple of ``(object, created)``, where ``object`` is the new
          or existing :py:class:`WorkRequest` and ``created`` is a boolean
          specifying whether a new object was created.
        """
        try:
            # Index: db_workrequest_wf_child_idx.
            return (
                self.work_request.children.unsuperseded().get(
                    task_type=TaskTypes.WORKFLOW,
                    task_name=task_name,
                    workflow_data_json__step=workflow_data.step,
                ),
                False,
            )
        except WorkRequest.DoesNotExist:
            return (
                self.work_request.create_child_workflow(
                    task_name=task_name,
                    task_data=task_data,
                    workflow_data=workflow_data,
                    relative_priority=relative_priority,
                ),
                True,
            )

    def work_request_ensure_child_signing(
        self,
        *,
        task_name: str,
        task_data: BaseTaskData | None = None,
        workflow_data: WorkRequestWorkflowData,
    ) -> tuple[WorkRequest, bool]:
        """
        Create a child SIGNING work request if one does not already exist.

        This assumes that ``workflow_data.step`` is unique among work
        request children with this task type and name, and that it is stable
        even if the workflow is orchestrated multiple times.

        :param task_name: the task name for the child work request.
        :param task_data: the task data for the child work request.
        :param workflow_data: the workflow data for the child work request.
        :param relative_priority: if creating a new work request, set the
          base priority of the child to the effective priority of this
          workflow plus this relative priority.
        :return: tuple of ``(object, created)``, where ``object`` is the new
          or existing :py:class:`WorkRequest` and ``created`` is a boolean
          specifying whether a new object was created.
        """
        try:
            # Index: db_workrequest_wf_child_idx.
            return (
                self.work_request.children.unsuperseded().get(
                    task_type=TaskTypes.SIGNING,
                    task_name=task_name,
                    workflow_data_json__step=workflow_data.step,
                ),
                False,
            )
        except WorkRequest.DoesNotExist:
            return (
                self.work_request.create_child_signing(
                    task_name=task_name,
                    task_data=task_data,
                    workflow_data=workflow_data,
                ),
                True,
            )

    def work_request_ensure_child_wait(
        self,
        *,
        task_name: str,
        task_data: BaseTaskData | None = None,
        workflow_data: WorkRequestWorkflowData,
    ) -> tuple[WorkRequest, bool]:
        """
        Create a child WAIT work request if one does not already exist.

        This assumes that ``workflow_data.step`` is unique among work
        request children with this task type and name, and that it is stable
        even if the workflow is orchestrated multiple times.

        :param task_name: the task name for the child work request.
        :param task_data: the task data for the child work request.
        :param workflow_data: the workflow data for the child work request.
        :return: tuple of ``(object, created)``, where ``object`` is the new
          or existing :py:class:`WorkRequest` and ``created`` is a boolean
          specifying whether a new object was created.
        """
        try:
            # Index: db_workrequest_wf_child_idx.
            return (
                self.work_request.children.unsuperseded().get(
                    task_type=TaskTypes.WAIT,
                    task_name=task_name,
                    workflow_data_json__step=workflow_data.step,
                ),
                False,
            )
        except WorkRequest.DoesNotExist:
            return (
                self.work_request.create_child_wait(
                    task_name=task_name,
                    task_data=task_data,
                    workflow_data=workflow_data,
                ),
                True,
            )

    def lookup_singleton_collection(
        self,
        category: CollectionCategory,
        *,
        workspace: Workspace | None = None,
    ) -> Collection:
        """Look up a singleton collection related to this workflow."""
        return self.work_request.lookup_single(
            f"_@{category}",
            workspace=workspace or self.workspace,
            expect_type=LookupChildType.COLLECTION,
        ).collection

    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return the list of input artifacts used by this workflow."""
        return []

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        """
        Return the list of artifact ids used by this workflow.

        .. deprecated:: 2025-12
           Use get_input_artifacts()
        """
        return [
            artifact_id
            for input_artifact in self.get_input_artifacts()
            for artifact_id in input_artifact.resolved_ids()
        ]

    @override
    def get_task_configuration_subject_context(
        self,
    ) -> tuple[str | None, str | None]:
        # Compute dynamic task to extract subject and configuration_context
        # TODO: these values can be computed in a separate methods if the rest
        # TODO: of compute_dynamic_data becomes too expensive to run twice
        early_dynamic_task_data = self.compute_dynamic_data()
        return (
            early_dynamic_task_data.subject,
            early_dynamic_task_data.configuration_context,
        )

    @override
    def execute(self) -> WorkRequest.Results:
        raise AssertionError("execute called on a Workflow task")

    @override
    def _execute(self) -> WorkRequestResults:
        raise AssertionError("_execute called on a Workflow task")


def describe_exc(exception: Exception) -> str:
    """User-facing description of exception."""
    message = str(exception)
    if not message:
        return repr(exception)
    return message


def _log_workflow_run_error(exc: WorkflowRunError) -> None:
    """
    Log a WorkflowRunError.

    This is in a separate function to make it easier to mock.
    """
    logger.warning(
        "Error running work request %s/%s (%s): %s",
        exc.work_request.task_type,
        exc.work_request.task_name,
        exc.work_request.id,
        exc.message,
        exc_info=True,
    )


def _finish_workflow_callback(
    work_request: WorkRequest, succeeded: bool
) -> None:
    """
    Handle the results of a workflow callback.

    Incremental workflow callbacks are set back to ``BLOCKED``.  Others are
    completed with a result depending on the value of ``succeeded``.
    """
    if (
        work_request.unblock_strategy == WorkRequest.UnblockStrategy.INCREMENTAL
        and work_request.dependencies.unterminated().exists()
    ):
        work_request.pending_at = None
        work_request.started_at = None
        work_request.status = WorkRequest.Statuses.BLOCKED
        work_request.save()
    else:
        work_request.mark_completed(
            WorkRequest.Results.SUCCESS
            if succeeded
            else WorkRequest.Results.FAILURE
        )


def orchestrate_workflow(work_request: WorkRequest) -> bool:
    """
    Orchestrate a workflow in whatever way is appropriate.

    For a workflow callback, run ``callback`` and mark the work request as
    completed.  For a workflow, run ``populate`` and unblock workflow
    children, but leave the workflow running until all its children have
    finished.  For any other work request, log an error.

    Return False if any errors were logged, otherwise True.
    """

    def run_or_raise_workflow_run_error[R: Any](
        func: Callable[[], R], message_template: str, code: str
    ) -> R:
        try:
            return func()
        except Exception as exc:
            message = describe_exc(exc)
            raise WorkflowRunError(
                work_request, message_template.format(message=message), code
            ) from exc

    try:
        run_or_raise_workflow_run_error(
            work_request.mark_running,
            "Failed to mark work request as running",
            "status-change-failed",
        )

        match (work_request.task_type, work_request.task_name):
            case (TaskTypes.INTERNAL, "workflow") | (TaskTypes.WORKFLOW, _):
                try:
                    orchestrator = work_request.get_task()
                except InternalTaskError as exc:
                    raise WorkflowRunError(
                        work_request, str(exc), "configure-failed"
                    ) from exc
                except TaskConfigError as exc:
                    raise WorkflowRunError(
                        work_request,
                        f"Failed to configure: {exc}",
                        "configure-failed",
                    ) from exc
                else:
                    assert isinstance(orchestrator, Workflow)

            case _:
                raise WorkflowRunError(
                    work_request,
                    "Does not have a workflow orchestrator",
                    "configure-failed",
                )

        match (work_request.task_type, work_request.task_name):
            case (TaskTypes.INTERNAL, "workflow"):
                parent = orchestrator.work_request
                # Workflow callbacks compute dynamic data for their parent
                # workflow.  If this fails, it's probably least confusing to
                # log information about both the workflow callback and the
                # workflow.
                run_or_raise_workflow_run_error(
                    orchestrator.ensure_dynamic_data,
                    f"Failed to compute dynamic data for "
                    f"{parent.task_type}/{parent.task_name} ({parent.id}): "
                    f"{{message}}",
                    "dynamic-data-failed",
                )

                succeeded = run_or_raise_workflow_run_error(
                    partial(orchestrator.callback, work_request),
                    "Orchestrator failed: {message}",
                    "orchestrator-failed",
                )

                _finish_workflow_callback(work_request, succeeded)

            case (TaskTypes.WORKFLOW, _):
                with work_request.running_workflow_orchestrator():
                    run_or_raise_workflow_run_error(
                        orchestrator.ensure_dynamic_data,
                        "Failed to compute dynamic data: {message}",
                        "dynamic-data-failed",
                    )

                    run_or_raise_workflow_run_error(
                        orchestrator.populate,
                        "Orchestrator failed: {message}",
                        "orchestrator-failed",
                    )

                    if work_request.children.exists():
                        orchestrator.work_request.unblock_workflow_children()

                work_request.maybe_finish_workflow()

                # The workflow is left running until all its children have
                # finished.

            case _:  # pragma: no cover
                # Already checked above.
                raise AssertionError(
                    f"Unexpected work request: "
                    f"{work_request.task_type}/{work_request.task_name}"
                )

        return True

    except WorkflowRunError as exc:
        _log_workflow_run_error(exc)
        try:
            work_request.mark_completed(
                WorkRequest.Results.ERROR,
                output_data=OutputData(
                    errors=[OutputDataError(message=exc.message, code=exc.code)]
                ),
            )
        except StatusChangeError:
            pass
        return False



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for SystemImageBuild class."""

from typing import override

import debusine.worker.tags as wtags
from debusine.tasks.models import BaseDynamicTaskData, DebootstrapVariant
from debusine.tasks.systemimagebuild import SystemImageBuild
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase


class SystemImageBuildImpl(SystemImageBuild):
    """Implementation of SystemImageBuild ontology."""

    @override
    def _cmdline(self) -> list[str]:
        return []  # pragma: no cover


class SystemImageBuildTests(
    ExternalTaskHelperMixin[SystemImageBuildImpl], TestCase
):
    """Tests SystemImageBuild class."""

    SAMPLE_TASK_DATA = {
        "bootstrap_options": {
            "architecture": "amd64",
            "extra_packages": ["hello", "python3"],
            "use_signed_by": True,
        },
        "bootstrap_repositories": [
            {
                "mirror": "https://deb.debian.org/debian",
                "suite": "stable",
                "components": ["main", "contrib"],
                "check_signature_with": "system",
            },
            {
                "types": ["deb-src"],
                "mirror": "https://example.com",
                "suite": "bullseye",
                "components": ["main"],
                "check_signature_with": "system",
                "keyring": {"url": "https://example.com/keyring.gpg"},
            },
        ],
        "disk_image": {
            "format": "raw",
            "partitions": [
                {
                    "size": 2,
                    "filesystem": "ext4",
                },
            ],
        },
    }

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:systemimagebuildimpl:version:1",
            ],
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        """Test compute_dynamic_data."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            BaseDynamicTaskData(
                subject="stable",
                parameter_summary="stable (amd64)",
                runtime_context="amd64:None:hello,python3",
                configuration_context="amd64",
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "stable")

    def test_get_configuration_context(self) -> None:
        self.assertEqual(self.get_configuration_context(), "amd64")

    def test_compute_dynamic_data_variant(self) -> None:
        task_data = self.SAMPLE_TASK_DATA.copy()
        task_data["bootstrap_options"]["variant"] = DebootstrapVariant.BUILDD
        self.configure_task(task_data=task_data)

        self.assertEqual(
            self.compute_dynamic_data(self.task).parameter_summary,
            "stable (amd64-buildd)",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for DebianArchiveManager."""

import datetime as dt
from pathlib import PurePath
from typing import ClassVar, override

from django.test import override_settings

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.db.models import Artifact, Collection, CollectionItem
from debusine.db.playground import scenarios
from debusine.server.collections import DebianArchiveManager, ItemAdditionError
from debusine.test.django import TestCase


@override_settings(LANGUAGE_CODE="en-us")
class DebianArchiveManagerTests(TestCase):
    """Tests for DebianArchiveManager."""

    scenario = scenarios.DefaultContext()
    collection: ClassVar[Collection]
    manager: ClassVar[DebianArchiveManager]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.collection = cls.scenario.workspace.get_singleton_collection(
            pc=cls.scenario.context_ui(), category=CollectionCategory.ARCHIVE
        )
        assert isinstance(cls.collection.manager, DebianArchiveManager)
        cls.manager = cls.collection.manager

    def create_source_package(
        self, name: str, version: str, paths: list[str]
    ) -> Artifact:
        """Create a minimal `debian:source-package` artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": name,
                "version": version,
                "type": "dpkg",
                "dsc_fields": {},
            },
            paths=paths,
            create_files=True,
            skip_add_files_in_store=True,
        )
        return artifact

    def create_binary_package(
        self,
        srcpkg_name: str,
        srcpkg_version: str,
        name: str,
        version: str,
        architecture: str,
        paths: list[str],
    ) -> Artifact:
        """Create a minimal `debian:binary-package` artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data={
                "srcpkg_name": srcpkg_name,
                "srcpkg_version": srcpkg_version,
                "deb_fields": {
                    "Package": name,
                    "Version": version,
                    "Architecture": architecture,
                },
                "deb_control_files": [],
            },
            paths=paths,
            create_files=True,
            skip_add_files_in_store=True,
        )
        return artifact

    def create_suite(self, name: str) -> Collection:
        """Create a minimal ``debian:suite`` collection."""
        return self.playground.create_collection(
            name, CollectionCategory.SUITE, workspace=self.scenario.workspace
        )

    def test_add_artifact_repository_index(self) -> None:
        """``add_artifact`` can add a repository index to the suite."""
        doc_artifact = self.playground.create_repository_index(
            "social-contract.txt", skip_add_files_in_store=True
        )

        collection_item = self.manager.add_artifact(
            doc_artifact,
            user=self.scenario.user,
            variables={"path": "doc/social-contract.txt"},
        )

        self.assertEqual(collection_item.name, "index:doc/social-contract.txt")
        self.assertEqual(
            collection_item.data, {"path": "doc/social-contract.txt"}
        )

    def test_add_artifact_repository_index_path_under_dists(self) -> None:
        """``add_artifact`` refuses paths not under ``dists/``."""
        artifact = self.playground.create_repository_index(
            "foo", skip_add_files_in_store=True
        )

        with self.assertRaisesRegex(
            ItemAdditionError,
            "Archive-level repository indexes may not have paths under dists/",
        ):
            self.manager.add_artifact(
                artifact,
                user=self.scenario.user,
                variables={"path": "dists/foo"},
            )

    def test_add_artifact_repository_index_violates_constraint(self) -> None:
        """``add_artifact`` reports repository index constraint violations."""
        artifacts = [
            self.playground.create_repository_index(
                "README", skip_add_files_in_store=True
            )
            for _ in range(2)
        ]
        self.manager.add_artifact(
            artifacts[0], user=self.scenario.user, variables={"path": "README"}
        )

        with self.assertRaisesRegex(
            ItemAdditionError,
            r'^duplicate key value violates unique constraint '
            r'"db_collectionitem_unique_active_name"',
        ):
            self.manager.add_artifact(
                artifacts[1],
                user=self.scenario.user,
                variables={"path": "README"},
            )

    def test_add_artifact_repository_index_replace(self) -> None:
        """``add_artifact`` can replace an existing repository index."""
        readme_artifact = self.playground.create_repository_index(
            "README", skip_add_files_in_store=True
        )
        collection_item = self.manager.add_artifact(
            readme_artifact,
            user=self.scenario.user,
            variables={"path": "README"},
        )
        readme_artifact2 = self.playground.create_repository_index(
            "README", skip_add_files_in_store=True
        )
        collection_item2 = self.manager.add_artifact(
            readme_artifact2,
            user=self.scenario.user,
            variables={"path": "README"},
            replace=True,
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.artifact, readme_artifact)
        self.assertEqual(collection_item.removed_by_user, self.scenario.user)
        self.assertIsNotNone(collection_item.removed_at)
        self.assertEqual(collection_item2.name, "index:README")
        self.assertEqual(collection_item2.artifact, readme_artifact2)
        self.assertIsNone(collection_item2.removed_at)

    def test_do_add_artifact_requires_path(self) -> None:
        """`do_add_artifact` requires a `path` variable."""
        readme_artifact = self.playground.create_repository_index(
            "README", skip_add_files_in_store=True
        )

        with self.assertRaisesRegex(
            ItemAdditionError,
            '^Adding debian:repository-index to debian:archive requires '
            '"path"$',
        ):
            self.manager.do_add_artifact(
                readme_artifact, user=self.scenario.user
            )

    def test_do_add_collection(self) -> None:
        """``do_add_collection`` can add a suite to the archive."""
        suite = self.create_suite("bookworm")

        collection_item = self.manager.add_collection(
            suite, user=self.scenario.user
        )

        self.assertEqual(collection_item.name, "bookworm")
        self.assertEqual(collection_item.data, {})

    def test_do_add_collection_copies_suite_constraints(self) -> None:
        """``do_add_collection`` copies pool-file constraints from the suite."""
        suite = self.create_suite("bookworm")
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        dsc_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.dsc"
        ).file
        tar_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.tar.xz"
        ).file
        suite_source_item = suite.manager.add_artifact(
            source_package_artifact,
            user=self.scenario.user,
            variables={"component": "main", "section": "devel"},
        )
        expected_constraints = [
            {
                "collection_item_id": suite_source_item.id,
                "constraint_name": "pool-file",
                "key": "pool/main/h/hello/hello_1.0.dsc",
                "value": f"sha256:{dsc_file.hash_digest.hex()}",
            },
            {
                "collection_item_id": suite_source_item.id,
                "constraint_name": "pool-file",
                "key": "pool/main/h/hello/hello_1.0.tar.xz",
                "value": f"sha256:{tar_file.hash_digest.hex()}",
            },
        ]
        self.assert_collection_has_constraints(suite, expected_constraints)
        self.assert_collection_has_constraints(self.collection, [])

        self.manager.add_collection(suite, user=self.scenario.user)

        for collection in (suite, self.collection):
            self.assert_collection_has_constraints(
                collection, expected_constraints
            )

    def test_do_add_collection_replaced_by_ignores_suite_constraints(
        self,
    ) -> None:
        new_suite = self.create_suite("bookworm")
        new_item = self.manager.add_collection(
            new_suite, user=self.scenario.user
        )

        old_suite = self.create_suite("bookworm-old")
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        dsc_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.dsc"
        ).file
        tar_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.tar.xz"
        ).file
        suite_source_item = old_suite.manager.add_artifact(
            source_package_artifact,
            user=self.scenario.user,
            variables={"component": "main", "section": "devel"},
        )
        expected_constraints = [
            {
                "collection_item_id": suite_source_item.id,
                "constraint_name": "pool-file",
                "key": "pool/main/h/hello/hello_1.0.dsc",
                "value": f"sha256:{dsc_file.hash_digest.hex()}",
            },
            {
                "collection_item_id": suite_source_item.id,
                "constraint_name": "pool-file",
                "key": "pool/main/h/hello/hello_1.0.tar.xz",
                "value": f"sha256:{tar_file.hash_digest.hex()}",
            },
        ]
        self.assert_collection_has_constraints(old_suite, expected_constraints)
        self.assert_collection_has_constraints(self.collection, [])

        self.manager.add_collection(
            old_suite, user=self.scenario.user, replaced_by=new_item
        )

        self.assert_collection_has_constraints(self.collection, [])

    def test_do_add_collection_different_workspace(self) -> None:
        other_workspace = self.playground.create_workspace(name="other")
        suite = self.playground.create_collection(
            "bookworm", CollectionCategory.SUITE, workspace=other_workspace
        )

        with self.assertRaisesRegex(
            ItemAdditionError,
            "Suite must be in the same workspace as the archive",
        ):
            self.manager.add_collection(suite, user=self.scenario.user)

    def test_do_add_collection_violates_constraint(self) -> None:
        """``add_collection`` reports constraint violations."""
        suites = [self.create_suite(name) for name in ("bookworm", "trixie")]
        # do_add_collection generates the item name from the collection
        # name, so this won't happen normally, but force it so that we can
        # test the constraint violation.
        CollectionItem.objects.create_from_collection(
            suites[0],
            parent_collection=self.collection,
            name="trixie",
            data={},
            created_by_user=self.scenario.user,
        )

        with self.assertRaisesRegex(
            ItemAdditionError,
            r'^duplicate key value violates unique constraint '
            r'"db_collectionitem_unique_active_name"',
        ):
            self.manager.add_collection(suites[1], user=self.scenario.user)

    def test_do_remove_item_artifact(self) -> None:
        """``do_remove_item`` removes an artifact."""
        readme_artifact = self.playground.create_repository_index(
            "README", skip_add_files_in_store=True
        )
        collection_item = self.manager.add_artifact(
            readme_artifact,
            user=self.scenario.user,
            variables={"path": "README"},
        )

        self.manager.remove_item(collection_item, user=self.scenario.user)

        self.assertEqual(collection_item.artifact, readme_artifact)
        self.assertEqual(collection_item.removed_by_user, self.scenario.user)
        self.assertIsInstance(collection_item.removed_at, dt.datetime)

    def test_do_remove_item_collection(self) -> None:
        """``do_remove_item`` can remove a suite from the archive."""
        suite = self.create_suite("bookworm")
        suite_item = self.manager.add_collection(suite, user=self.scenario.user)

        self.manager.remove_item(suite_item, user=self.scenario.user)

        self.assertEqual(suite_item.collection, suite)
        self.assertEqual(suite_item.removed_by_user, self.scenario.user)
        self.assertIsInstance(suite_item.removed_at, dt.datetime)

    def test_do_remove_item_collection_retains_suite_constraints(self) -> None:
        """If `may_reuse_versions=False`, versions may not be reused."""
        suite = self.create_suite("bookworm")
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        dsc_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.dsc"
        ).file
        tar_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.tar.xz"
        ).file
        suite_source_item = suite.manager.add_artifact(
            source_package_artifact,
            user=self.scenario.user,
            variables={"component": "main", "section": "devel"},
        )
        suite_item = self.manager.add_collection(suite, user=self.scenario.user)

        self.manager.remove_item(suite_item, user=self.scenario.user)

        # Since `may_reuse_versions` defaults to False, the constraints on
        # matching pool file contents are retained.
        self.assert_collection_has_constraints(
            self.collection,
            [
                {
                    "collection_item_id": suite_source_item.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/hello_1.0.dsc",
                    "value": f"sha256:{dsc_file.hash_digest.hex()}",
                },
                {
                    "collection_item_id": suite_source_item.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/hello_1.0.tar.xz",
                    "value": f"sha256:{tar_file.hash_digest.hex()}",
                },
            ],
        )

    def test_do_remove_item_collection_reuse_versions(self) -> None:
        """If `may_reuse_versions=True`, versions may be reused."""
        self.collection.data["may_reuse_versions"] = True
        suite = self.create_suite("bookworm")
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        suite.manager.add_artifact(
            source_package_artifact,
            user=self.scenario.user,
            variables={"component": "main", "section": "devel"},
        )
        suite_item = self.manager.add_collection(suite, user=self.scenario.user)

        self.manager.remove_item(suite_item, user=self.scenario.user)

        self.assert_collection_has_constraints(self.collection, [])

    def test_lookup_unexpected_format_raise_lookup_error(self) -> None:
        """`lookup` raises `LookupError`: invalid format."""
        with self.assertRaisesRegex(
            LookupError, '^Unexpected lookup format: "binary:hello"$'
        ):
            self.manager.lookup("binary:hello")

    def test_lookup_return_none(self) -> None:
        """`lookup` returns None if there are no matches."""
        self.assertIsNone(self.manager.lookup("source:hello"))
        self.assertIsNone(self.manager.lookup("binary:hello_amd64"))
        self.assertIsNone(self.manager.lookup("index:missing"))
        self.assertIsNone(self.manager.lookup("name:nonexistent"))

    def test_lookup_source(self) -> None:
        """`lookup` returns a matching source package item from a suite."""
        suites = [self.create_suite(name) for name in ("bookworm", "trixie")]
        for suite in suites:
            self.manager.add_collection(suite, user=self.scenario.user)

        items: list[CollectionItem] = []
        for suite, name, version in (
            (suites[0], "hello", "1.0"),
            (suites[1], "hello", "1.1"),
            (suites[0], "base-files", "1.0"),
        ):
            source_package_artifact = self.create_source_package(
                name, version, []
            )
            items.append(
                suite.manager.add_artifact(
                    source_package_artifact,
                    user=self.scenario.user,
                    variables={"component": "main", "section": "devel"},
                )
            )

        source_package_artifact = self.create_source_package("hello", "1.2", [])
        items.append(
            suites[1].manager.add_artifact(
                source_package_artifact,
                user=self.scenario.user,
                variables={"component": "main", "section": "devel"},
            )
        )
        suites[1].manager.remove_item(items[-1], user=self.scenario.user)

        # items[0] (bookworm/hello_1.0) and items[1] (trixie/hello_1.1) both
        # match, but items[1] has a higher version.  items[3]
        # (trixie/hello_1.2) has been removed.
        self.assertEqual(self.manager.lookup("source:hello"), items[1])

        self.assertEqual(
            self.manager.lookup("source-version:hello_1.0"), items[0]
        )
        self.assertEqual(
            self.manager.lookup("source-version:hello_1.1"), items[1]
        )
        self.assertIsNone(self.manager.lookup("source-version:hello_1.2"))

    def test_lookup_binary(self) -> None:
        """`lookup` returns a matching binary package item."""
        suites = [self.create_suite(name) for name in ("bookworm", "trixie")]
        for suite in suites:
            self.manager.add_collection(suite, user=self.scenario.user)

        items: list[CollectionItem] = []
        for suite, srcpkg_name, srcpkg_version, name, version, architecture in (
            (suites[0], "hello", "1.0", "libhello1", "1.0", "amd64"),
            (suites[0], "hello", "1.0", "libhello1", "1.0", "s390x"),
            (suites[0], "hello", "1.0", "libhello-doc", "1.0", "all"),
            (suites[1], "hello", "1.1", "libhello1", "1.1", "amd64"),
            (suites[0], "base-files", "1.0", "base-files", "1:1.0", "amd64"),
        ):
            binary_package_artifact = self.create_binary_package(
                srcpkg_name, srcpkg_version, name, version, architecture, []
            )
            items.append(
                suite.manager.add_artifact(
                    binary_package_artifact,
                    user=self.scenario.user,
                    variables={
                        "component": "main",
                        "section": "devel",
                        "priority": "optional",
                    },
                )
            )

        binary_package_artifact = self.create_binary_package(
            "hello", "1.2", "libhello1", "1.2", "amd64", []
        )
        items.append(
            suites[1].manager.add_artifact(
                binary_package_artifact,
                user=self.scenario.user,
                variables={
                    "component": "main",
                    "section": "devel",
                    "priority": "optional",
                },
            )
        )
        suites[1].manager.remove_item(items[-1], user=self.scenario.user)

        # items[0] (bookworm/libhello1_1.0_amd64) and items[3]
        # (trixie/libhello_1.1_amd64) both match, but items[3] has a higher
        # version.  items[5] (trixie/libhello1_1.2_amd64) has been removed.
        self.assertEqual(
            self.manager.lookup("binary:libhello1_amd64"), items[3]
        )

        self.assertEqual(
            self.manager.lookup("binary-version:libhello1_1.0_amd64"), items[0]
        )
        self.assertEqual(
            self.manager.lookup("binary-version:libhello1_1.0_s390x"), items[1]
        )
        self.assertEqual(
            self.manager.lookup("binary-version:libhello-doc_1.0_all"), items[2]
        )
        self.assertEqual(
            self.manager.lookup("binary-version:libhello-doc_1.0_amd64"),
            items[2],
        )
        self.assertEqual(
            self.manager.lookup("binary-version:libhello-doc_1.0_s390x"),
            items[2],
        )
        self.assertEqual(
            self.manager.lookup("binary-version:libhello1_1.1_amd64"), items[3]
        )
        self.assertIsNone(
            self.manager.lookup("binary-version:libhello1_1.1_s390x")
        )
        self.assertIsNone(
            self.manager.lookup("binary-version:libhello1_1.2_amd64")
        )

    def test_lookup_repository_index(self) -> None:
        """``lookup`` returns a matching repository index."""
        suites = [self.create_suite(name) for name in ("bookworm", "trixie")]
        for suite in suites:
            self.manager.add_collection(suite, user=self.scenario.user)

        items: list[CollectionItem] = []
        for suite, path in (
            (suites[0], "main/source/Sources.xz"),
            (suites[0], "main/binary-amd64/Packages.xz"),
            (suites[0], "Release"),
            (suites[1], "main/source/Sources.xz"),
            (suites[1], "main/binary-i386/Packages.xz"),
            (suites[1], "Release"),
        ):
            index_artifact = self.playground.create_repository_index(
                PurePath(path).name, skip_add_files_in_store=True
            )
            items.append(
                suite.manager.add_artifact(
                    index_artifact,
                    user=self.scenario.user,
                    variables={"path": path},
                )
            )

        index_artifact = self.playground.create_repository_index(
            "Packages.xz", skip_add_files_in_store=True
        )
        items.append(
            suites[1].manager.add_artifact(
                index_artifact,
                user=self.scenario.user,
                variables={"path": "main/binary-amd64/Packages.xz"},
                replace=True,
            )
        )

        readme_artifact = self.playground.create_repository_index(
            "README", skip_add_files_in_store=True
        )
        items.append(
            self.manager.add_artifact(
                readme_artifact,
                user=self.scenario.user,
                variables={"path": "README"},
            )
        )

        self.assertEqual(
            self.manager.lookup("index:dists/bookworm/main/source/Sources.xz"),
            items[0],
        )
        self.assertEqual(
            self.manager.lookup(
                "index:dists/bookworm/main/binary-amd64/Packages.xz"
            ),
            items[1],
        )
        self.assertEqual(
            self.manager.lookup("index:dists/bookworm/Release"), items[2]
        )
        self.assertEqual(
            self.manager.lookup("index:dists/trixie/main/source/Sources.xz"),
            items[3],
        )
        # items[4] and items[6] would both match, but only items[6] is
        # active.
        self.assertEqual(
            self.manager.lookup(
                "index:dists/trixie/main/binary-amd64/Packages.xz"
            ),
            items[6],
        )
        self.assertEqual(
            self.manager.lookup("index:dists/trixie/Release"), items[5]
        )
        self.assertEqual(self.manager.lookup("index:README"), items[7])
        self.assertIsNone(self.manager.lookup("index:missing"))
        self.assertIsNone(self.manager.lookup("index:dists/bookworm/missing"))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
r"""
Collection of tasks.

The debusine.tasks module hierarchy hosts a collection of
:py:class:`BaseExternalTask` that are used by workers to fulfill
:py:class:`debusine.db.models.WorkRequest`\ s sent by the debusine
scheduler.

Creating a new task requires adding a new file containing a class inheriting
from the :py:class:`BaseExternalTask` or :py:class:`RunCommandTask` base class.
The name of the class must be unique among all child classes.

A child class must, at the very least, override the
:py:meth:`BaseExternalTask.execute` method.
"""

import datetime as dt
import inspect
import logging
import os
import shlex
import shutil
import signal
import subprocess
import tempfile
import traceback
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from collections.abc import Collection, Generator
from contextlib import contextmanager
from pathlib import Path
from typing import (
    Any,
    AnyStr,
    BinaryIO,
    ClassVar,
    IO,
    TYPE_CHECKING,
    TextIO,
    Union,
    overload,
    override,
)

from debian.debian_support import Version

from debusine.artifacts import WorkRequestDebugLogs
from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    TaskTypes,
    WorkRequestResults,
)
from debusine.client.debusine import Debusine
from debusine.client.models import ArtifactResponse, RelationType
from debusine.tasks.executors import (
    ExecutorInterface,
    InstanceInterface,
    executor_class,
)
from debusine.tasks.inputs import (
    BuildArchitectureInput,
    ExtraRepositoriesInput,
    Stage,
    TaskFieldContainer,
)
from debusine.tasks.models import (
    BackendType,
    BaseDynamicTaskData,
    BaseDynamicTaskDataWithExecutor,
    BaseDynamicTaskDataWithExtraRepositories,
    BaseTaskData,
    BaseTaskDataWithExecutor,
    BaseTaskDataWithExtraRepositories,
    WorkerType,
)
from debusine.utils import extract_generic_type_arguments

if TYPE_CHECKING:
    from _typeshed import OpenBinaryModeWriting, OpenTextModeWriting

    import debusine.tasks.tags as ttags
    import debusine.worker.tags as wtags


#: Fallback arhitecturer to use when a task doesn't explicitly specify one
FALLBACK_ARCHITECTURE = "amd64"


class TaskConfigError(Exception):
    """Exception raised when there is an issue with a task configuration."""

    @override
    def __init__(
        self, message: str | None, original_exception: Exception | None = None
    ) -> None:
        """
        Initialize the TaskConfigError.

        :param message: human-readable message describing the error.
        :param original_exception: the exception that triggered this error,
          if applicable. This is used to provide additional information.
        """
        super().__init__(message)
        self.original_exception = original_exception

    def add_parent_message(self, msg: str) -> None:
        """Prepend the error message with one from the containing scope."""
        if self.args[0]:
            self.args = (f"{msg}: {self.args[0]}",)
        else:
            self.args = (msg,)

    @override
    def __str__(self) -> str:
        """
        Return a string representation.

        If an original exception is present, its representation is appended
        to the message for additional context.
        """
        if self.original_exception:
            if self.args[0] is not None:
                return (
                    f"{self.args[0]} (Original exception: "
                    f"{self.original_exception})"
                )
            else:
                return str(self.original_exception)
        else:
            return str(self.args[0])


def ensure_artifact_categories(
    *,
    configuration_key: str,
    category: str,
    expected: Collection[ArtifactCategory],
) -> None:
    """
    Validate that the artifact's category is one of the expected categories.

    :param configuration_key: Optional key to identify the source of the
      artifact. Provides additional context in error messages.
    :param category: The category to validate.
    :param expected: A collection of valid artifact categories.
    """
    valid_categories = [cat.value for cat in sorted(expected)]
    if category not in valid_categories:
        raise TaskConfigError(
            f"{configuration_key}: unexpected artifact "
            f"category: '{category}'. "
            f"Valid categories: {valid_categories}"
        )


def ensure_collection_category(
    *, configuration_key: str, category: str, expected: CollectionCategory
) -> None:
    """
    Validate that the collection's category is as expected.

    :param configuration_key: Optional key to identify the source of the
      artifact. Provides additional context in error messages.
    :param category: The category to validate.
    :param expected: The expected collection category.
    """
    if category != expected:
        raise TaskConfigError(
            f"{configuration_key}: unexpected collection category: "
            f"'{category}'. Expected: '{expected}'"
        )


def analyze_external_worker_tasks() -> dict[str, Any]:
    """
    Return dictionary with metadata for each BaseExternalTask.

    Subclasses of BaseExternalTask get registered in
    BaseExternalTask._sub_tasks. Return a dictionary with the metadata of each
    of the subtasks.

    This method is executed in the worker when submitting the dynamic
    metadata.
    """
    metadata = {}

    for registry in BaseExternalTask._sub_tasks.values():
        for task_class in registry.values():
            metadata.update(task_class.analyze_worker())

    return metadata


def get_provided_external_worker_tags() -> set[str]:
    """
    Return the set of worker tags provided by all available tasks.

    This method is called on the worker to collect information about the
    worker.
    """
    tags: set[str] = set()

    for registry in BaseExternalTask._sub_tasks.values():
        for task_class in registry.values():
            tags |= task_class.get_provided_worker_tags()

    return tags


class BaseExternalTask[
    TD: BaseTaskData,
    DTD: BaseDynamicTaskData = BaseDynamicTaskData,
](TaskFieldContainer, metaclass=ABCMeta):
    r"""
    Base class for external tasks.

    A BaseExternalTask object serves two purposes: encapsulating the logic of
    what needs to be done to execute the task on an external worker (cf
    :py:meth:`configure` and :py:meth:`execute`, that are run on the worker),
    and supporting the server's task-specific logic, such as scheduling and UI
    presentation.

    Scheduling support is done in a two-step process: collating metadata from
    each worker (with the :py:meth:`analyze_worker` method that is run on a
    worker) and then, based on this metadata, see if a task is suitable (with
    :py:meth:`can_run_on` that is executed on the scheduler).

    If tag-based scheduling is in use, that is replaced by
    :py:meth`get_provided_worker_tags`, which runs on the worker and reports
    provided worker tags that need to match what tasks require.

    Concrete subclasses must implement:

    * ``run(execute_directory: Path) -> bool``: Do the main work of the
      task.

    Most concrete subclasses should also implement:

    * ``fetch_input(self, destination) -> bool``. Download the needed
      artifacts into destination. Suggestion: can use
      ``fetch_artifact(artifact, dir)`` to download them.
      (default: return True)
    * ``configure_for_execution(self, download_directory: Path) -> bool``
      (default: return True)
    * ``check_directory_for_consistency_errors(self, build_directory: Path)
      -> list[str]``
      (default: return an empty list, indicating no errors)
    * ``upload_artifacts(self, directory: Path, \*, execution_success: bool)``.
      The member variable self._source_artifacts_ids is set by
      ``fetch_input()`` and can be used to create the relations between
      uploaded artifacts and downloaded artifacts.
      (default: return True)

    Most concrete task implementations should inherit from
    :py:class:`RunCommandTask` instead.
    """

    #: Class used as the in-memory representation of task data.
    task_data_type: type[TD]
    data: TD

    #: Class used as the in-memory representation of dynamic task data.
    dynamic_task_data_type: type[DTD]
    dynamic_data: DTD | None

    #: Must be overridden by child classes to document the current version of
    #: the task's code. A task will only be scheduled on a worker if its task
    #: version is the same as the one running on the scheduler.
    TASK_VERSION: int | None = None

    #: The worker type must be suitable for the task type.  TaskTypes.WORKER
    #: requires an external worker; TaskTypes.SERVER requires a Celery
    #: worker; TaskTypes.SIGNING requires a signing worker.
    TASK_TYPE: TaskTypes = TaskTypes.WORKER

    name: ClassVar[str]
    _sub_tasks: dict[
        TaskTypes, dict[str, type["BaseExternalTask['Any', 'Any']"]]
    ] = defaultdict(dict)

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Register the subclass into BaseExternalTask._sub_tasks.

        Used by BaseExternalTask.class_from_name() to return the class given
        the name.
        """
        super().__init_subclass__(**kwargs)

        # The name of the task. It is computed by converting the class name
        # to lowercase.
        cls.name = getattr(cls, "TASK_NAME", cls.__name__.lower())

        # The task data types, computed by introspecting the type arguments
        # used to specialize this generic class.
        [
            cls.task_data_type,
            cls.dynamic_task_data_type,
        ] = extract_generic_type_arguments(cls, BaseExternalTask)

        if inspect.isabstract(cls):
            # Don't list abstract base classes as tasks.
            return

        registry = cls._sub_tasks[cls.TASK_TYPE]

        # The same sub-task could register twice
        # (but assert that is the *same* class, not a different
        # subtask with a name with a different capitalisation)
        if cls.name in registry and registry[cls.name] != cls:
            raise AssertionError(f'Two Tasks with the same name: {cls.name!r}')

        # Make sure SERVER and WORKER do not have conflicting task names
        match cls.TASK_TYPE:
            case TaskTypes.SERVER:
                if cls.name in cls._sub_tasks[TaskTypes.WORKER]:
                    raise AssertionError(
                        f'{cls.name!r} already registered as a Worker task'
                    )
            case TaskTypes.WORKER:
                if cls.name in cls._sub_tasks[TaskTypes.SERVER]:
                    raise AssertionError(
                        f'{cls.name!r} already registered as a Server task'
                    )

        registry[cls.name] = cls

    def __init__(
        self,
        task_data: dict[str, Any],
        dynamic_task_data: dict[str, Any] | None = None,
        *,
        wrapped: bool = False,
    ) -> None:
        """
        Initialize the task.

        :param wrapped: set to True when this task is managed by a wrapping task
        """
        super().__init__()

        #: A :py:class:`logging.Logger` instance that can be used in child
        #: classes when you override methods to implement the task.
        self.logger = logging.getLogger("debusine.tasks")

        #: Validated task data submitted through :py:meth:`configure` without
        # BaseExternalTask generic data
        self._configure(task_data, dynamic_task_data)

        self.working_directory: Path | None = None
        self.worker_environment_variables: dict[str, str] = {}

        # Task is aborted: the task does not need to be executed, and can be
        # stopped if it is already running
        self._aborted = False

        self.work_request_id: int | None = None

        # Injectable HTTP timeout from the worker configuration
        self.http_timeout: int | tuple[int, int] | None = None

        # Workspace is used when downloading and uploading artifacts.
        # When the worker instantiates the task it should set
        # self.workspace_name (see issue #186).
        self.workspace_name: str | None = None

        # fetch_input() add the downloaded artifacts. Used by
        # `BaseExternalTask._upload_work_request_debug_logs()` and maybe by
        # required method `upload_artifacts()`.
        #
        # This is distinct from get_input_artifacts_ids, which is used to
        # extract IDs from dynamic_data for use by UI views
        self._source_artifacts_ids: list[int] = []

        self.debusine: Debusine | None = None

        self.executor: ExecutorInterface | None = None
        self.executor_instance: InstanceInterface | None = None

        # Restore values of task input fields stored in dynamic data
        if not wrapped and self.dynamic_data is not None:
            self.restore_field_values(self.dynamic_data)

        self.post_init()

    def post_init(self) -> None:
        """Specific post-init code."""

    def set_working_directory(self, working_directory: Path) -> None:
        working_directory.mkdir(parents=True, exist_ok=True)
        self.working_directory = working_directory

    def set_worker_environment_variables(self, env: dict[str, str]) -> None:
        self.worker_environment_variables = env

    @property
    def log_directory(self) -> Path:
        """Return the directory storing logs of the task's execution."""
        assert self.working_directory
        return self.working_directory / "logs"

    def append_to_log_file(self, filename: str, lines: list[str]) -> None:
        """
        Open log file and write contents into it.

        :param filename: use self.open_debug_log_file(filename)
        :param lines: write contents to the logfile
        """
        with self.open_debug_log_file(filename) as file:
            file.writelines([line + "\n" for line in lines])

    @overload
    def open_debug_log_file(
        self, filename: str, *, mode: "OpenTextModeWriting" = "a"
    ) -> TextIO: ...

    @overload
    def open_debug_log_file(
        self, filename: str, *, mode: "OpenBinaryModeWriting"
    ) -> BinaryIO: ...

    def open_debug_log_file(
        self,
        filename: str,
        *,
        mode: Union["OpenTextModeWriting", "OpenBinaryModeWriting"] = "a",
    ) -> IO[Any]:
        """
        Open a temporary file and return it.

        The files are always for the same temporary directory, calling it twice
        with the same file name will open the same file.

        The caller must call .close() when finished writing.
        """
        debug_file = self.log_directory / filename
        debug_file.parent.mkdir(exist_ok=True)
        return debug_file.open(mode)

    @classmethod
    def prefix_with_task_name(cls, text: str) -> str:
        """:return: the ``text`` prefixed with the task name and a colon."""
        if cls.TASK_TYPE is TaskTypes.WORKER:
            # Worker tasks are left unprefixed for compatibility
            return f"{cls.name}:{text}"
        else:
            return f"{cls.TASK_TYPE.lower()}:{cls.name}:{text}"

    @classmethod
    def analyze_worker(cls) -> dict[str, Any]:
        """
        Return dynamic metadata about the current worker.

        This method is called on the worker to collect information about the
        worker. The information is stored as a set of key-value pairs in a
        dictionary.

        That information is then reused on the scheduler to be fed to
        :py:meth:`can_run_on` and determine if a task is suitable to be
        executed on the worker.

        Derived objects can extend the behaviour by overriding
        the method, calling ``metadata = super().analyze_worker()``,
        and then adding supplementary data in the dictionary.

        To avoid conflicts on the names of the keys used by different tasks
        you should use key names obtained with
        ``self.prefix_with_task_name(...)``.

        :return: a dictionary describing the worker.
        :rtype: dict.
        """
        version_key_name = cls.prefix_with_task_name("version")
        return {
            version_key_name: cls.TASK_VERSION,
        }

    @classmethod
    def get_provided_worker_tags(cls) -> set[str]:
        """
        Return the set of worker tags provided by this task class.

        This method is called on the worker to collect information about the
        worker.
        """
        # Prevent circular import
        import debusine.worker.tags as wtags

        return {
            wtags.TASK_PREFIX
            + f"{cls.TASK_TYPE.lower()}:{cls.name}:version:{cls.TASK_VERSION}"
        }

    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        """
        Check if the specified worker can run the task.

        This method shall take its decision solely based on the supplied
        ``worker_metadata`` and on the configured task data (``self.data``).

        The default implementation always returns True unless
        :py:attr:`TASK_TYPE` doesn't match the worker type or there's a
        mismatch between the :py:attr:`TASK_VERSION` on the scheduler side
        and on the worker side.

        Derived objects can implement further checks by overriding the method
        in the following way::

            if not super().can_run_on(worker_metadata):
                return False

            if ...:
                return False

            return True

        :param dict worker_metadata: The metadata collected from the worker by
            running :py:meth:`analyze_worker` on all the tasks on the worker
            under consideration.
        :return: the boolean result of the check.
        :rtype: bool.
        """
        worker_type = worker_metadata.get("system:worker_type")
        if (self.TASK_TYPE, worker_type) not in {
            (TaskTypes.WORKER, WorkerType.EXTERNAL),
            (TaskTypes.SERVER, WorkerType.CELERY),
            (TaskTypes.SIGNING, WorkerType.SIGNING),
        }:
            return False

        version_key_name = self.prefix_with_task_name("version")

        if worker_metadata.get(version_key_name) != self.TASK_VERSION:
            return False

        return True

    def compute_scheduler_tags(
        self,
    ) -> tuple["ttags.TaskProvidedTags", "wtags.TaskRequiredTags"]:
        """
        Compute task-provided and task-required tagsets.

        This is called twice, and both times with only the ``stage=PENDING``
        fields resolved:

        * Once before task configuration, to compute the early tag set used for
          the configuration lookup
        * Once after task configuration, to the final tag set used for
          scheduling
        """
        import debusine.tasks.tags as ttags
        import debusine.worker.tags as wtags

        provided = ttags.TaskProvidedTags()
        required = wtags.TaskRequiredTags()

        return provided, required

    @abstractmethod
    def get_subject(self) -> str | None:
        """
        Return the task subject used to look up task configuration.

        This method can only be called after all the task fields have been
        resolved.
        """

    def get_configuration_context(self) -> str | None:
        """
        Return the configuration context used to look up task configuration.

        This method can only be called after all the task fields have been
        resolved.
        """
        return None

    def compute_dynamic_data(self) -> DTD:
        """
        Build a dynamic task data structure for this task.

        This method can only be called after all the task fields have been
        resolved.

        :returns: the newly created dynamic task data
        """
        return self.dynamic_task_data_type()

    def _configure(
        self,
        task_data: dict[str, Any],
        dynamic_task_data: dict[str, Any] | None = None,
    ) -> None:
        """
        Configure the task with the supplied ``task_data``.

        The supplied data is validated against the pydantic data model for the
        type of task being configured. If validation fails, a TaskConfigError
        is raised. Otherwise, the `data` attribute is filled with the supplied
        `task_data`.

        :param dict task_data: The supplied data describing the task.
        :raises TaskConfigError: if the dict does not validate.
        """
        try:
            self.data = self.task_data_type(**task_data)
            self.dynamic_data = (
                None
                if dynamic_task_data is None
                else self.dynamic_task_data_type(**dynamic_task_data)
            )
        except ValueError as exc:
            raise TaskConfigError(None, original_exception=exc)

    def execute_logging_exceptions(self) -> WorkRequestResults:
        """Execute self.execute() logging any raised exceptions."""
        try:
            return self.execute()
        except Exception as exc:
            self.logger.info("Exception in Task %s", self.name, exc_info=True)
            raise exc

    def execute(self) -> WorkRequestResults:
        """
        Call the _execute() method, upload debug artifacts.

        See _execute() for more information.

        :return: result of the _execute() method.
        """  # noqa: D402
        result = self._execute()

        self._upload_work_request_debug_logs()

        return result

    def abort(self) -> None:
        """Task does not need to be executed. Once aborted cannot be changed."""
        self._aborted = True

    @property
    def aborted(self) -> bool:
        """
        Return if the task is aborted.

        Tasks cannot transition from aborted -> not-aborted.
        """
        return self._aborted

    @staticmethod
    def class_from_name(
        task_type: TaskTypes, task_name: str
    ) -> type["BaseExternalTask['Any', 'Any']"]:
        """
        Return class for :param task_name (case-insensitive).

        :param task_type: type of task to look up

        __init_subclass__() registers BaseExternalTask subclasses' into
        BaseExternalTask._sub_tasks.
        """
        if (registry := BaseExternalTask._sub_tasks.get(task_type)) is None:
            raise ValueError(f"{task_type!r} is not a registered task type")

        task_name_lowercase = task_name.lower()
        if (cls := registry.get(task_name_lowercase)) is None:
            raise ValueError(
                f"{task_name_lowercase!r} is not a registered"
                f" {task_type} task_name"
            )

        return cls

    def get_input_artifacts_ids(self) -> list[int]:
        """
        Return the list of input artifact IDs used by this task.

        This refers to the artifacts actually used by the task. If
        dynamic_data is empty, this returns the empty list.

        This is used by views to show what artifacts were used by a task.
        `_source_artifacts_ids` cannot be used for this purpose because it is
        only set during task execution.
        """
        artifact_ids: list[int] = []

        if self.dynamic_data is not None:
            artifact_ids.extend(self.dynamic_data.get_input_artifacts_ids())

        for name, field in self.inputs.items():
            if name not in self.__dict__:
                continue
            artifact_ids.extend(field.get_artifact_ids(self.__dict__[name]))

        return artifact_ids

    def configure_server_access(self, debusine: Debusine) -> None:
        """Set the object to access the server."""
        self.debusine = debusine

    @contextmanager
    def _temporary_directory(self, prefix: str) -> Generator[Path]:
        assert self.working_directory
        with tempfile.TemporaryDirectory(
            prefix=f"{prefix}-",
            dir=self.working_directory,
        ) as directory:
            yield Path(directory)

    def _log_exception(self, exc: Exception, stage: str) -> None:
        """Log an unexpected exception into the task log for stage."""
        exc_type = type(exc).__name__
        exc_traceback = traceback.format_exc()

        log_message = [
            f"Exception type: {exc_type}",
            f"Message: {exc}",
            "",
            exc_traceback,
        ]

        self.append_to_log_file(f"{stage}.log", log_message)

    def fetch_artifact(
        self, artifact_id: int, destination: Path
    ) -> ArtifactResponse:
        """
        Download artifact_id to destination.

        Add artifact_id to self._source_artifacts_ids.
        """
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        if not self.workspace_name:  # Required for cross-scope artifacts
            raise AssertionError("self.workspace_name not set")

        artifact_response = self.debusine.download_artifact(
            artifact_id,
            destination,
            tarball=False,
            workspace=self.workspace_name,
        )
        self._source_artifacts_ids.append(artifact_id)
        return artifact_response

    def fetch_input(self, destination: Path) -> bool:  # noqa: ARG002, U100
        """
        Download artifacts needed by the task, update self.source_artifacts_ids.

        Task might use self.data.input to download the relevant artifacts.

        The method self.fetch_artifact(artifact, destination) might be used
        to download the relevant artifacts and update
        self.source_artifacts_ids.
        """
        return True

    def configure_for_execution(
        self,
        download_directory: Path,  # noqa: ARG002, U100
    ) -> bool:
        """
        Configure task: set variables needed for the self._cmdline().

        Called after the files are downloaded via `fetch_input()`.
        """
        return True

    def prepare_to_run(
        self,
        download_directory: Path,  # noqa: ARG002, U100
        execute_directory: Path,  # noqa: ARG002, U100
    ) -> None:
        """
        Prepare the execution environment to do the main work of the task.

        :param download_directory: Directory where required artifacts were
            previously downloaded by self.fetch_input(download_directory).
        :param execute_directory: Directory where the task is expected to
            do its job.
        """
        # Do nothing by default.

    @abstractmethod
    def run(
        self,
        execute_directory: Path,  # noqa: ARG002, U100
    ) -> WorkRequestResults:
        """
        Do the main work of the task.

        :param execute_directory: Temporary directory prepared by the worker
            and by former steps of the task (prepare_to_run notably) where the
            task is expected to do its job.
        """

    def check_directory_for_consistency_errors(
        self,
        build_directory: Path,  # noqa: ARG002, U100
    ) -> list[str]:
        """Return list of errors after doing the main work of the task."""
        return []

    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,  # noqa: U100
    ) -> None:
        """Upload the artifacts for the task."""
        # Do nothing by default.

    def cleanup(self) -> None:
        """Clean up after running the task."""
        # Do nothing by default.

    def _execute(self) -> WorkRequestResults:  # noqa: C901
        """
        Execute the requested task.

        The task must first have been configured. It is allowed to take
        as much time as required. This method will only be run on a worker. It
        is thus allowed to access resources local to the worker.

        It is recommended to fail early by raising a :py:exc:`TaskConfigError`
        if the parameters of the task let you anticipate that it has no chance
        of completing successfully.

        Fetch the required input, execute the command and upload artifacts.

        Flow is:

        - Call ``self.fetch_input():`` download the artifacts
        - Call ``self.configure_for_execution()``: to set any member variables
          that might be used in ``self._cmdline()``
        - Call ``self.prepare_to_run()`` to prepare the execution
          environment to do the main work of the task
        - Call ``self.run()`` to do the main work of the task
        - Call ``self.check_directory_for_consistency_errors()``. If any
          errors are returned save them into consistency.log.
        - Call ``self.upload_artifacts(exec_dir, execution_success=succeeded)``.
        - Return ``execution_result``  (the Worker will set the result)

        :return: SUCCESS to indicate success, FAILURE for a failure, ERROR for
            an internal error, SKIPPED if the task turned out to be a noop.
        :raises TaskConfigError: if the parameters of the work request are
            incompatible with the worker.
        """
        with (
            self._temporary_directory("exec") as execute_directory,
            self._temporary_directory("fetch") as download_directory,
            self.open_debug_log_file("stages.log") as stages_log,
        ):

            def log_stage(message: str) -> None:
                """Log a stage, both to the worker log and to a WR debug log."""
                self.logger.info(
                    "Work request %s: %s", self.work_request_id, message
                )
                print(
                    f"{dt.datetime.now(dt.UTC).isoformat()} {message}",
                    file=stages_log,
                )

            try:
                try:
                    log_stage("Fetching input")
                    if not self.fetch_input(download_directory):
                        return WorkRequestResults.FAILURE
                except Exception as exc:
                    self._log_exception(exc, stage="fetch_input")
                    return WorkRequestResults.FAILURE

                try:
                    log_stage("Configuring for execution")
                    if not self.configure_for_execution(download_directory):
                        return WorkRequestResults.FAILURE
                except Exception as exc:
                    self._log_exception(exc, stage="configure_for_execution")
                    return WorkRequestResults.FAILURE

                try:
                    log_stage("Preparing to run")
                    self.prepare_to_run(download_directory, execute_directory)

                    log_stage("Running")

                    execution_result = self.run(execute_directory)
                except Exception as exc:
                    self._log_exception(exc, stage="execution")
                    return WorkRequestResults.FAILURE

                try:
                    log_stage("Checking output")
                    if errors := self.check_directory_for_consistency_errors(
                        execute_directory
                    ):
                        self.append_to_log_file(
                            "consistency.log", sorted(errors)
                        )
                        return WorkRequestResults.FAILURE

                    log_stage("Uploading artifacts")
                    self.upload_artifacts(
                        execute_directory, execution_result=execution_result
                    )
                except Exception as exc:
                    self._log_exception(exc, stage="post_execution")
                    return WorkRequestResults.FAILURE
            finally:
                try:
                    log_stage("Cleaning up")
                    self.cleanup()
                except Exception as exc:
                    self._log_exception(exc, stage="post_execution")
                    execution_result = WorkRequestResults.FAILURE

        if execution_result == WorkRequestResults.SUCCESS:
            # We mainly care about stages.log for failures.  It's rather
            # boring for successes.
            (self.log_directory / "stages.log").unlink()

        return execution_result

    def _upload_work_request_debug_logs(self) -> None:
        """
        Create a WorkRequestDebugLogs artifact and upload the logs.

        The logs might exist in self.log_directory and were
        added via self.open_debug_log_file() or self.append_debug_log_file().

        For each self._source_artifacts_ids: create a relation from
        WorkRequestDebugLogs to source_artifact_id.
        """
        if not self.log_directory.exists():
            return

        work_request_debug_logs_artifact = WorkRequestDebugLogs.create(
            files=self.log_directory.glob("*")
        )

        assert self.debusine
        remote_artifact = self.debusine.upload_artifact(
            work_request_debug_logs_artifact,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )

        for source_artifact_id in self._source_artifacts_ids:
            self.debusine.relation_create(
                remote_artifact.id,
                source_artifact_id,
                RelationType.RELATES_TO,
            )

        shutil.rmtree(self.log_directory)


class BaseTaskWithExecutor[
    TDE: BaseTaskDataWithExecutor,
    DTDE: BaseDynamicTaskDataWithExecutor,
](BaseExternalTask[TDE, DTDE], metaclass=ABCMeta):
    r"""
    Base for tasks with executor capabilities.

    Concrete subclasses must implement ``fetch_input()``,
    ``configure_for_execution()``, ``run()``,
    ``check_directory_for_consistency_errors()``, and
    ``upload_artifacts()``, as documented by :py:class:`BaseExternalTask`.
    """

    DEFAULT_BACKEND = BackendType.UNSHARE

    #: Task build architecture
    build_architecture = BuildArchitectureInput(
        stage=Stage.PENDING,
        verbose_name="Build architecture",
        help_text="System architecture for the build environment",
    )

    @override
    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        # Also require a matching build architecture
        if not super().can_run_on(worker_metadata):
            return False

        if self.build_architecture not in worker_metadata.get(
            "system:architectures", []
        ):
            return False

        return True

    def _prepare_executor(self) -> None:
        """
        Prepare the executor.

        * Set self.executor to the new executor with self.backend and the
          appropriate environment ID (which must have been looked up by the
          task's compute_dynamic_data method)
        * Download the image
        """
        assert self.debusine
        assert self.dynamic_data
        assert self.dynamic_data.environment_id
        assert self.workspace_name
        assert self.working_directory
        self.executor = executor_class(self.backend)(
            self.debusine,
            self.dynamic_data.environment_id,
            workspace=self.workspace_name,
            working_directory=self.working_directory,
        )
        self.executor.download_image()

    def _prepare_executor_instance(self) -> None:
        """
        Create and start an executor instance.

        If self.executor is None: call self._prepare_executor()

        Set self.executor_instance to the new executor instance, starts
        the instance.
        """
        if self.executor is None:
            self._prepare_executor()

        assert self.executor
        self.executor_instance = self.executor.create()
        self.executor_instance.start()

    @property
    def backend(self) -> BackendType:
        """Return the backend name to use."""
        backend = self.data.backend
        if backend == BackendType.AUTO:
            backend = self.DEFAULT_BACKEND
        return backend

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple["ttags.TaskProvidedTags", "wtags.TaskRequiredTags"]:
        # Prevent circular import
        import debusine.worker.tags as wtags

        provided, required = super().compute_scheduler_tags()
        required.add(
            wtags.ProvenanceRequired.SYSTEM,
            {
                wtags.EXECUTOR_PREFIX + self.backend,
                wtags.BUILD_ARCH_PREFIX + self.build_architecture,
            },
        )
        return provided, required

    def get_package_version(self, package_name: str) -> Version:
        """
        Get the installed version of a package in the executor.

        :raises subprocess.CalledProcessError: if the package is not
          installed (also logged to ``install.log``).
        """
        assert self.executor_instance is not None
        cmd = ["dpkg-query", "--show", "--showformat=${Version}", package_name]
        try:
            return Version(
                self.executor_instance.run(
                    cmd, check=True, run_as_root=True, text=True
                ).stdout
            )
        except subprocess.CalledProcessError as e:
            with self.open_debug_log_file("install.log", mode="a") as log:
                log.write(
                    f"Failed to determine available {package_name} version.\n"
                    f"command: {shlex.join(cmd)}\n"
                    f"exitcode: {e.returncode}\n"
                    f"stdout: {e.stdout}\n"
                    f"stderr: {e.stderr}\n"
                )
                raise

    @override
    def prepare_to_run(
        self, download_directory: Path, execute_directory: Path
    ) -> None:
        """Copy the download and execution directories into the executor."""
        if self.executor_instance:
            self.executor_instance.create_user()
            non_root_user = self.executor_instance.non_root_user
            self.executor_instance.mkdir(
                download_directory.parent,
                user=non_root_user,
                group=non_root_user,
                parents=True,
            )
            self.executor_instance.directory_push(
                download_directory,
                download_directory.parent,
                user=non_root_user,
                group=non_root_user,
            )
            self.executor_instance.mkdir(
                execute_directory,
                user=non_root_user,
                group=non_root_user,
                parents=True,
            )

    def run_executor_command(
        self,
        cmd: list[str],
        log_filename: str,
        run_as_root: bool = False,
        check: bool = True,
    ) -> None:
        """Run cmd within the executor, logging the output to log_name."""
        assert self.executor_instance
        with self.open_debug_log_file(log_filename, mode="ab") as cmd_log:
            msg = f"Executing: {shlex.join(cmd)}\n"
            cmd_log.write(msg.encode())
            cmd_log.flush()
            p = self.executor_instance.run(
                cmd,
                run_as_root=run_as_root,
                stdout=cmd_log,
                stderr=cmd_log,
                check=check,
            )
            msg = f"Execution completed (exit code {p.returncode})\n"
            cmd_log.write(msg.encode())

    @override
    def cleanup(self) -> None:
        """
        Clean up after running the task.

        Some tasks use the executor in upload_artifacts, so we clean up the
        executor here rather than in run().
        """
        if self.executor_instance and self.executor_instance.is_started():
            self.executor_instance.stop()


class RunCommandTask[
    TD: BaseTaskData,
    DTD: BaseDynamicTaskData = BaseDynamicTaskData,
](BaseExternalTask[TD, DTD], metaclass=ABCMeta):
    r"""
    A :py:class:`BaseExternalTask` that executes commands and uploads artifacts.

    Concrete subclasses must implement:

    * ``_cmdline(self) -> list[str]``
    * ``task_result(self, returncode: Optional[int], execute_directory: Path)
      -> WorkRequestResults`` (defaults to SUCCESS)

    They must also implement ``configure_for_execution()``,
    ``fetch_input()``, ``check_directory_for_consistency_errors()``, and
    ``upload_artifacts()``, as documented by :py:class:`BaseExternalTask`.
    (They do not need to implement ``run()``, but may do so if they need to
    run multiple commands rather than just one.)

    Use ``self.append_to_log_file()`` / ``self.open_debug_log_file()`` to
    provide information for the user (it will be available to the user as an
    artifact).

    Command execution uses process groups to make sure that the command and
    possible spawned commands are finished, and cancels the execution of the
    command if ``BaseExternalTask.aborted()`` is True. If the task also
    inherits from BaseTaskWithExecutor, then the command is run inside
    an executor instance.

    Optionally: _cmdline_as_root() and _cmd_env() may be implemented, to
    customize behaviour.

    See the main entry point ``BaseExternalTask._execute()`` for details of the
    flow.
    """

    # If CAPTURE_OUTPUT_FILENAME is not None: self.run_via_executor()
    # creates a file in the cwd of the command to save the stdout. The file
    # is available in self.upload_artifacts().
    CAPTURE_OUTPUT_FILENAME: str | None = None

    CMD_LOG_SEPARATOR = "--------------------"
    CMD_LOG_FILENAME = "cmd-output.log"

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:
        """
        Run the command returned by ``self._cmdline()``.

        The command is executed with ``execute_directory`` as the current
        directory. That directory is a dedicated temporary directory setup
        by the worker for use by the task.

        When RunCommandTask is combined with BaseTaskWithExecutor, then
        the command is executed inside the executor instance.

        .. note::

          If the member variable CAPTURE_OUTPUT_FILENAME is set:
          create a file with its name with the stdout of the command. Otherwise,
          the stdout of the command is saved in self.CMD_LOG_FILENAME together
          with stderr's output.
        """
        cmd = self._cmdline()
        self.logger.info("Executing: %s", " ".join(cmd))

        returncode = self.run_cmd(
            cmd,
            execute_directory,
            env=self._cmd_env(),
            run_as_root=self._cmdline_as_root(),
            capture_stdout_filename=self.CAPTURE_OUTPUT_FILENAME,
        )

        self.logger.info("%s exited with code %s", cmd[0], returncode)

        return self.task_result(returncode, execute_directory)

    def task_result(
        self,
        returncode: int | None,
        execute_directory: Path,  # noqa: ARG002, U100
    ) -> WorkRequestResults:
        """
        Sub-tasks can evaluate if the task was a success or failure.

        By default, return SUCCESS if returncode is 0, FAILURE otherwise.
        Sub-classes can re-implement it.

        :param returncode: return code of the command, or None if aborted
        :param execute_directory: directory with the output of the task
        :return: WorkRequestResults.
        """
        if returncode == 0:
            return WorkRequestResults.SUCCESS
        return WorkRequestResults.FAILURE

    @abstractmethod
    def _cmdline(self) -> list[str]:
        """Return the command to execute, as a list of program arguments."""

    def _cmd_env(self) -> dict[str, str]:
        """Return the (minimal) environment to execute the command under."""
        env = {
            "PATH": os.environ["PATH"],
        }
        env.update(self.worker_environment_variables)
        return env

    @staticmethod
    def _cmdline_as_root() -> bool:
        """Run _cmdline() as root."""
        return False

    @staticmethod
    def _write_utf8(file: BinaryIO, text: str) -> None:
        file.write(text.encode("utf-8", errors="replace") + b"\n")

    def _write_popen_result(
        self, file: BinaryIO, p: subprocess.Popen[AnyStr]
    ) -> None:
        self._write_utf8(file, f"\naborted: {self.aborted}")
        self._write_utf8(file, f"returncode: {p.returncode}")

    def run_cmd(
        self,
        cmd: list[str],
        working_directory: Path,
        *,
        env: dict[str, str] | None = None,
        run_as_root: bool = False,
        capture_stdout_filename: str | None = None,
    ) -> int | None:
        """
        Run cmd in working_directory. Create self.CMD_OUTPUT_FILE log file.

        If BaseExternalTask.aborted == True terminates the process.

        :param cmd: command to execute with its arguments.
        :param working_directory: working directory where the command
          is executed.
        :param run_as_root: if True, run the command as root. Otherwise,
          the command runs as the worker's user
        :param capture_stdout_filename: for some commands the output of the
          command is the output of stdout (e.g. lintian) and not a set of files
          generated by the command (e.g. sbuild). If capture_stdout is not None,
          save the stdout into this file. The caller can then use it.
        :return: returncode of the process or None if aborted
        """
        if env is None:
            env = self._cmd_env()

        with self.open_debug_log_file(
            self.CMD_LOG_FILENAME, mode="ab"
        ) as cmd_log:
            self._write_utf8(cmd_log, f"cmd: {shlex.join(cmd)}")

            out_file: BinaryIO
            if capture_stdout_filename:
                self._write_utf8(
                    cmd_log,
                    "output (contains stderr only, stdout was captured):",
                )
                capture_stdout = working_directory / capture_stdout_filename
                out_file = capture_stdout.open(mode="wb")
            else:
                self._write_utf8(
                    cmd_log, "output (contains stdout and stderr):"
                )
                out_file = cmd_log

            try:
                return self._run_cmd(
                    cmd,
                    working_directory,
                    env=env,
                    run_as_root=run_as_root,
                    cmd_log=cmd_log,
                    out_file=out_file,
                )
            finally:
                file_names = "\n".join(
                    str(file.relative_to(working_directory))
                    for file in sorted(working_directory.rglob("*"))
                )

                self._write_utf8(cmd_log, "\nFiles in working directory:")
                self._write_utf8(cmd_log, file_names)

                self._write_utf8(cmd_log, self.CMD_LOG_SEPARATOR)
                if capture_stdout_filename:
                    cmd_log.write(capture_stdout.read_bytes() + b"\n")
                    self._write_utf8(cmd_log, self.CMD_LOG_SEPARATOR)

                out_file.close()

    def _run_cmd(
        self,
        cmd: list[str],
        working_directory: Path,
        *,
        env: dict[str, str],
        run_as_root: bool,
        cmd_log: BinaryIO,
        out_file: BinaryIO,
    ) -> int | None:
        """
        Execute cmd.

        :param run_as_root: if using an executor: run command as root or user.
          If not using an executor and run_as_root is True: raise ValueError().
        :param cmd_log: save the command log (parameters, stderr, return code,
          stdout)
        :param out_file: save the command stdout (might be the same as
          cmd_log or a different file)

        :return: returncode of the process or None if aborted
        """
        # Need to flush or subprocess.Popen() overwrites part of it
        cmd_log.flush()
        out_file.flush()

        run_kwargs: dict[str, Any] = {
            "cwd": working_directory,
            "env": env,
            "stderr": cmd_log.fileno(),
            "stdout": out_file.fileno(),
        }

        if self.executor_instance:
            return self.executor_instance.run(
                cmd, run_as_root=run_as_root, **run_kwargs
            ).returncode

        if run_as_root:
            raise ValueError("run_as_root requires an executor")

        assert env is not None
        with self._temporary_directory("tmp") as tmpdir:
            tmpdir.chmod(0o755)  # Some tasks will root containers inside here
            env["TMPDIR"] = str(tmpdir)

            p = subprocess.Popen(cmd, start_new_session=True, **run_kwargs)

            process_group = os.getpgid(p.pid)

            while True:
                if self.aborted:
                    break

                try:
                    self._wait_popen(p, timeout=1)
                    break
                except subprocess.TimeoutExpired:
                    pass

        if self.aborted:
            self.logger.debug("Task (cmd: %s PID %s) aborted", cmd, p.pid)
            try:
                if not self._send_signal_pid(p.pid, signal.SIGTERM):
                    # _send_signal_pid failed probably because cmd finished
                    # after aborting and before sending the signal
                    #
                    # p.poll() to read the returncode and avoid leaving cmd
                    # as zombie
                    p.poll()

                    # Kill possible processes launched by cmd
                    self._send_signal_group(process_group, signal.SIGKILL)
                    self.logger.debug("Could not send SIGTERM to %s", p.pid)

                    self._write_popen_result(cmd_log, p)
                    return None

                # _wait_popen with a timeout=5 to leave 5 seconds of grace
                # for the cmd to finish after sending SIGTERM
                self._wait_popen(p, timeout=5)
            except subprocess.TimeoutExpired:
                # SIGTERM was sent and 5 seconds later cmd
                # was still running. A SIGKILL to the process group will
                # be sent
                self.logger.debug(
                    "Task PID %s not finished after SIGTERM", p.pid
                )

            # debusine sends a SIGKILL if:
            # - SIGTERM was sent to cmd AND cmd was running 5 seconds later:
            #   SIGTERM was not enough so SIGKILL to the group is needed
            # - SIGTERM was sent to cmd AND cmd finished: SIGKILL to the
            #   group to make sure that there are not processes spawned
            #   by cmd running
            # (note that a cmd could launch processes in a new group
            # could be left running)
            self._send_signal_group(process_group, signal.SIGKILL)
            self.logger.debug("Sent SIGKILL to process group %s", process_group)

            # p.poll() to set p.returncode and avoid leaving cmd
            # as a zombie process.
            # But cmd might be left as a zombie process: if cmd was in a
            # non-interruptable kernel call p.returncode will be None even
            # after p.poll() and it will be left as a zombie process
            # (until debusine worker dies and the zombie is adopted by
            # init and waited on by init). If this happened there we might be a
            # ResourceWarning from Popen.__del__:
            # "subprocess %s is still running"
            #
            # A solution would e to wait (p.waitpid()) that the
            # process finished dying. This is implemented in the unit test
            # to avoid the warning but not implemented here to not delay
            # the possible shut down of debusine worker
            p.poll()
            self._write_popen_result(cmd_log, p)
            self.logger.debug("Returncode for PID %s: %s", p.pid, p.returncode)

            return None
        else:
            # The cmd has finished. The cmd might have spawned
            # other processes. debusine will kill any alive processes.
            #
            # If they existed they should have been finished by cmd:
            # run_cmd() should not leave processes behind.
            #
            # Since the parent died they are adopted by init and on
            # killing them they are not zombie.
            # (cmd might have spawned new processes in a different process
            # group: if this is the case they will be left running)
            self._send_signal_group(process_group, signal.SIGKILL)

        self._write_popen_result(cmd_log, p)

        return p.returncode

    def _wait_popen(
        self, popen: subprocess.Popen[AnyStr], timeout: float
    ) -> int:
        return popen.wait(timeout)

    @staticmethod
    def _send_signal_pid(pid: int, signal: signal.Signals) -> bool:
        try:
            os.kill(pid, signal)
        except ProcessLookupError:
            return False

        return True

    @staticmethod
    def _send_signal_group(process_group: int, signal: signal.Signals) -> None:
        """Send signal to the process group."""
        try:
            os.killpg(process_group, signal)
        except ProcessLookupError:
            pass


class ExtraRepositoryMixin[
    TDER: BaseTaskDataWithExtraRepositories,
    DTDER: BaseDynamicTaskDataWithExtraRepositories,
](BaseExternalTask[TDER, DTDER], metaclass=ABCMeta):
    """Methods for configuring external APT repositories."""

    data: TDER

    #: Extra repositories configured in task data
    extra_repositories = ExtraRepositoriesInput()

    # Deb822 style sources:
    extra_repository_sources: list[Path]
    # One-line sources:
    extra_repository_lists: list[Path]
    # GPG keys (for apt < 2.3.10):
    extra_repository_keys: list[Path]

    @override
    def post_init(self) -> None:
        super().post_init()
        self.extra_repository_sources = []
        self.extra_repository_lists = []
        self.extra_repository_keys = []

    def supports_inline_signed_by(self, codename: str) -> bool:
        """Determine if codename supports inline signatures."""
        # apt < 2.3.10 has no support for keys inline in Signed-By
        return codename not in (
            "stretch",
            "buster",
            "bullseye",
            "xenial",
            "bionic",
            "focal",
        )

    def write_extra_repository_config(
        self, codename: str, destination: Path
    ) -> None:
        """
        Write ``extra_repositories`` config to files in destination.

        ``extra_repository_keys`` will be populated with keys to install
        into /etc/apt/keyrings, if needed.

        ``extra_repository_sources`` will be populated with deb822 sources
        files.
        """
        assert self.dynamic_data is not None

        inline_signed_by = self.supports_inline_signed_by(codename)
        build_arch = self.data.build_architecture
        filtered_repos = [
            repo
            for repo in (self.dynamic_data.extra_repositories or [])
            if repo.architectures is None or (build_arch in repo.architectures)
        ]
        for i, extra_repo in enumerate(filtered_repos):
            signed_by = None
            if not inline_signed_by:
                if extra_repo.signing_key:
                    path = destination / f"extra_apt_key_{i}.asc"
                    path.write_text(extra_repo.signing_key + "\n")
                    self.extra_repository_keys.append(path)
                    signed_by = f"/etc/apt/keyrings/{path.name}"

            path = destination / f"extra_repository_{i}.sources"
            path.write_text(
                extra_repo.as_deb822_source(signed_by_filename=signed_by)
            )
            self.extra_repository_sources.append(path)



# Generated by Django 4.2.28 on 2026-04-28 10:25

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0051_collectionrelation'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionrelation',
            name='source',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='relations',
                to='db.collection',
            ),
        ),
        migrations.AlterField(
            model_name='collectionrelation',
            name='target',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='targeted_by',
                to='db.collection',
            ),
        ),
        migrations.AlterField(
            model_name='collectionrelation',
            name='type',
            field=models.CharField(
                choices=[
                    ('suite-forked-from', 'Forked from'),
                    ('suite-based-on', 'Based on'),
                    ('suite-requires', 'Requires'),
                    ('suite-targeting', 'Targeting'),
                    ('suite-default-qa-results', 'Default QA results'),
                ],
                max_length=32,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
View extension to add context-dependent UI shortcut widgets.

A UI shortcut is a widget that can be rendered associated to a UI element to
provide a shortcut to commonly used views or actions related to it.
"""

from dataclasses import dataclass
from typing import override

from django.template.backends.utils import csrf_input
from django.template.context import BaseContext
from django.urls import reverse
from django.utils.html import format_html

from debusine.artifacts.models import ArtifactCategory
from debusine.db.models import (
    Artifact,
    CollectionItem,
    FileInArtifact,
    WorkRequest,
)
from debusine.db.models.artifacts import (
    ARTIFACT_CATEGORY_ICON_NAMES,
    ARTIFACT_CATEGORY_SHORT_NAMES,
)
from debusine.db.models.collections import CollectionRelation
from debusine.web.icons import Icons
from debusine.web.views.base import Widget


@dataclass(kw_only=True)
class UIShortcut(Widget):
    """Renderable UI shortcut."""

    #: User-readable label
    label: str
    #: Icon (name in the Bootstrap icon set, without the leading "bi-")
    icon: str
    #: Target URL for the action
    url: str

    @override
    def render(self, context: BaseContext) -> str:  # noqa: ARG002, U100
        """Render the shortcut as an <a> button."""
        return format_html(
            "<a class='btn btn-outline-secondary'"
            " href='{url}' title='{label}'>"
            "<span class='bi bi-{icon}'></span>"
            "</a>",
            label=self.label,
            icon=self.icon,
            url=self.url,
        )


@dataclass(kw_only=True)
class UIShortcutPOST(UIShortcut):
    """UI shortcut that triggers a POST."""

    #: Button style (from
    #: https://getbootstrap.com/docs/5.0/components/buttons/, without the
    #: leading "btn-")
    button_style: str = "primary"

    @override
    def render(self, context: BaseContext) -> str:
        """Render the shortcut as a form."""
        return format_html(
            "<form method='post' action='{url}'>{csrf}"
            "<button type='submit' class='btn btn-{button_style} bi bi-{icon}'"
            " title='{label}'></button>"
            "</form>",
            csrf=csrf_input(context["request"]),
            label=self.label,
            icon=self.icon,
            url=self.url,
            button_style=self.button_style,
        )


def create_work_request_retry(work_request: WorkRequest) -> UIShortcutPOST:
    """Create a shortcut to retry a work request."""
    return UIShortcutPOST(
        label="Retry work request",
        icon=Icons.WORK_REQUEST_RETRY,
        url=work_request.get_absolute_url_retry(),
    )


def create_work_request_abort(work_request: WorkRequest) -> UIShortcutPOST:
    """Create a shortcut to abort a work request."""
    return UIShortcutPOST(
        label="Abort work request",
        icon=Icons.WORK_REQUEST_ABORT,
        url=work_request.get_absolute_url_abort(),
        button_style="danger",
    )


def create_artifact_view(artifact: Artifact) -> UIShortcut:
    """Create a shortcut to view an artifact."""
    category = ArtifactCategory(artifact.category)
    short_name = ARTIFACT_CATEGORY_SHORT_NAMES.get(category, "artifact")
    return UIShortcut(
        label=f"View {short_name} artifact",
        icon=ARTIFACT_CATEGORY_ICON_NAMES.get(category, "folder"),
        url=artifact.get_absolute_url(),
    )


def create_file_view(file_in_artifact: FileInArtifact) -> UIShortcut:
    """Create a shortcut to view a file."""
    return UIShortcut(
        label=f"View {file_in_artifact.path}",
        icon=Icons.FILE_VIEW,
        url=file_in_artifact.get_absolute_url(),
    )


def create_file_view_raw(file_in_artifact: FileInArtifact) -> UIShortcut:
    """Create a shortcut to stream a file inline."""
    return UIShortcut(
        label=f"View {file_in_artifact.path} raw",
        icon=Icons.FILE_VIEW_RAW,
        url=file_in_artifact.get_absolute_url_raw(),
    )


def create_file_download(file_in_artifact: FileInArtifact) -> UIShortcut:
    """Create a shortcut to download a file."""
    return UIShortcut(
        label=f"Download {file_in_artifact.path}",
        icon=Icons.FILE_DOWNLOAD,
        url=file_in_artifact.get_absolute_url_download(),
    )


def create_artifact_download(artifact: Artifact) -> UIShortcut:
    """Create a shortcut to download an artifact."""
    return UIShortcut(
        label="Download artifact",
        icon=Icons.ARTIFACT_DOWNLOAD,
        url=artifact.get_absolute_url_download(),
    )


def create_collection_item(item: CollectionItem) -> UIShortcut:
    """Create a shortcut to view collection item details."""
    return UIShortcut(
        label="Collection item details",
        icon=Icons.COLLECTION_ITEM_DETAILS,
        url=item.get_absolute_url(),
    )


def create_collection_relation_delete(item: CollectionRelation) -> UIShortcut:
    """Create a shortcut to delete a CollectionRelation."""
    return UIShortcut(
        label="Delete collection relation",
        icon=Icons.DELETE,
        url=reverse(
            "workspaces:collections:collection_relations:delete",
            kwargs={
                "wname": item.source.workspace.name,
                "ccat": item.source.category,
                "cname": item.source.name,
                "pk": item.pk,
            },
        ),
    )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the Debsign task."""

import os
import shutil
import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, override
from unittest import mock

from django.test import TestCase as DjangoTestCase
from django.test import TransactionTestCase, override_settings
from nacl.public import PrivateKey

import debusine.worker.tags as wtags
from debusine.artifacts import (
    LocalArtifact,
    SourcePackage,
    Upload,
    WorkRequestDebugLogs,
)
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianUpload,
    EmptyArtifactData,
    WorkRequestResults,
)
from debusine.assets.models import AssetCategory
from debusine.client.models import (
    ArtifactResponse,
    AssetPermissionCheckResponse,
    RemoteArtifact,
)
from debusine.signing.db.models import Key
from debusine.signing.gnupg import gpg_ephemeral_context
from debusine.signing.models import SigningMode
from debusine.signing.tasks import Debsign
from debusine.signing.tasks.models import DebsignDynamicData
from debusine.signing.tasks.tests.helpers import SigningTaskTestCase
from debusine.tasks import FALLBACK_ARCHITECTURE
from debusine.tasks.models import WorkerType
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_remote_artifact,
    create_work_request_response,
)
from debusine.utils import read_changes


class DebsignTests(SigningTaskTestCase[Debsign], TestCase, DjangoTestCase):
    """Unit tests for the :py:class:`Debsign` task."""

    SAMPLE_TASK_DATA = {"unsigned": 2, "key": "ABC123"}

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set_field_artifact(
            "unsigned",
            id=2,
            category=ArtifactCategory.UPLOAD,
            data=DebianUpload(
                type="dpkg",
                changes_fields={
                    "Architecture": "source",
                    "Files": [{"name": "hello_1.0.dsc"}],
                },
            ),
        )

    def test_get_provided_worker_tags(self) -> None:
        self.mock_is_command_available({"debsign": True})
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "signing:debsign:version:1",
                wtags.CAP_PREFIX + "debsign",
            ],
        )

    def test_get_provided_worker_tags_debsign_not_available(self) -> None:
        self.mock_is_command_available({"debsign": False})
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [wtags.TASK_PREFIX + "signing:debsign:version:1"],
        )

    def test_analyze_worker(self) -> None:
        """Test the analyze_worker() method."""
        self.mock_is_command_available({"debsign": True})
        metadata = self.task.analyze_worker()
        self.assertEqual(metadata["signing:debsign:available"], True)

    def test_analyze_worker_debsign_not_available(self) -> None:
        """analyze_worker() handles debsign not being available."""
        self.mock_is_command_available({"debsign": False})
        metadata = self.task.analyze_worker()
        self.assertEqual(metadata["signing:debsign:available"], False)

    def test_can_run_on(self) -> None:
        """can_run_on returns True if debsign is available."""
        self.assertTrue(
            self.task.can_run_on(
                {
                    "system:worker_type": WorkerType.SIGNING,
                    "signing:debsign:available": True,
                    "signing:debsign:version": self.task.TASK_VERSION,
                    "system:architectures": [FALLBACK_ARCHITECTURE],
                }
            )
        )

    def test_can_run_on_mismatched_task_version(self) -> None:
        """can_run_on returns False for mismatched task versions."""
        self.assertFalse(
            self.task.can_run_on(
                {
                    "system:worker_type": WorkerType.SIGNING,
                    "signing:debsign:available": True,
                    "signing:debsign:version": self.task.TASK_VERSION + 1,
                }
            )
        )

    def test_can_run_on_missing_tool(self) -> None:
        """can_run_on returns False if debsign is not available."""
        self.assertFalse(
            self.task.can_run_on(
                {
                    "system:worker_type": WorkerType.SIGNING,
                    "signing:debsign:available": False,
                    "signing:debsign:version": self.task.TASK_VERSION,
                }
            )
        )

    def test_can_run_on_not_available(self) -> None:
        """can_run_on returns False if debsign is not available."""
        self.assertFalse(
            self.task.can_run_on(
                {
                    # "signing:debsign:available" is not set
                    "system:worker_type": WorkerType.SIGNING,
                    "signing:debsign:version": self.task.TASK_VERSION,
                    "system:architectures": [FALLBACK_ARCHITECTURE],
                }
            )
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            require_system={wtags.CAP_PREFIX + "debsign"}
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            DebsignDynamicData(unsigned_id=2),
        )

    def test_get_subject(self) -> None:
        self.assertIsNone(self.get_subject())

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = DebsignDynamicData(unsigned_id=1)
        self.assertEqual(self.task.get_input_artifacts_ids(), [1])

    def test_fetch_input_wrong_category(self) -> None:
        """fetch_input checks the category of the unsigned artifact."""
        temp_path = self.create_temporary_directory()
        dsc_path = temp_path / "foo.dsc"
        self.write_dsc_file(dsc_path, [])
        source_package = SourcePackage.create(
            name="foo", version="1.0", files=[dsc_path]
        )
        source_package_id = 1
        source_package_response = self.make_artifact_response(
            source_package, source_package_id
        )

        task = Debsign(
            task_data={"unsigned": source_package_id, "key": "ABC123"},
            dynamic_task_data={"unsigned_id": source_package_id},
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = source_package_response
        destination = self.create_temporary_directory()

        self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            "Expected unsigned artifact of category debian:upload; got "
            "debian:source-package\n",
        )
        debusine_mock.download_artifact.assert_called_once_with(
            source_package_id,
            destination,
            tarball=mock.ANY,
            workspace=task.workspace_name,
        )

    def test_fetch_input_missing_key(self) -> None:
        """fetch_input requires the signing key to exist in the database."""
        fingerprint = "0" * 64
        directory = self.create_temporary_directory()
        changes_file = directory / "foo.changes"
        self.write_changes_file(changes_file, [])
        upload = Upload.create(changes_file=changes_file)
        upload_id = 2
        upload_response = self.make_artifact_response(upload, upload_id)

        task = Debsign(
            task_data={"unsigned": upload_id, "key": fingerprint},
            dynamic_task_data={"unsigned_id": upload_id},
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = upload_response
        destination = self.create_temporary_directory()

        self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            f"Signing key openpgp:{fingerprint} does not exist\n",
        )
        debusine_mock.download_artifact.assert_called_once_with(
            upload_id,
            destination,
            tarball=mock.ANY,
            workspace=task.workspace_name,
        )

    def test_fetch_input_permission_denied(self) -> None:
        key = Key.objects.create(
            purpose=Key.Purpose.OPENPGP_UPLOAD,
            fingerprint="0" * 64,
            private_key={},
            public_key=b"",
        )
        directory = self.create_temporary_directory()
        changes_file = directory / "foo.changes"
        self.write_changes_file(changes_file, [])
        upload = Upload.create(changes_file=changes_file)
        upload_id = 2
        upload_response = self.make_artifact_response(upload, upload_id)

        task = Debsign(
            task_data={"unsigned": upload_id, "key": key.fingerprint},
            dynamic_task_data={"unsigned_id": upload_id},
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = upload_response
        debusine_mock.asset_permission_check.return_value = (
            AssetPermissionCheckResponse(
                has_permission=False,
                username="testuser",
                user_id=123,
                resource={},
            )
        )
        destination = self.create_temporary_directory()

        self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            f"User testuser does not have permission to sign "
            f"{upload_id} with openpgp:{key.fingerprint}.\n",
        )
        debusine_mock.download_artifact.assert_called_once_with(
            upload_id,
            destination,
            tarball=mock.ANY,
            workspace=task.workspace_name,
        )

    def test_update_changes_file(self) -> None:
        directory = self.create_temporary_directory()
        changes_file = directory / "foo.changes"
        self.write_changes_file(changes_file, [])
        self.task._files = self.make_files_response(
            {"foo.changes": changes_file}
        )

        debusine_mock = self.mock_debusine(self.task)
        debusine_mock.base_api_url = "https://example.com/"
        debusine_mock.work_request_get.return_value = (
            create_work_request_response(id=100, workflow_root=99)
        )

        self.task.update_changes_file(directory)
        updated_changes = read_changes(changes_file)

        self.assertIn("Debusine-User", updated_changes)
        self.assertEqual(
            updated_changes["Debusine-User"], "https://example.com/-/users/123/"
        )
        self.assertIn("Debusine-Workflow", updated_changes)
        self.assertEqual(
            updated_changes["Debusine-Workflow"],
            "https://example.com/debusine/Testing/work-request/99/",
        )
        self.assertIn("Debusine-Work-Request", updated_changes)
        self.assertEqual(
            updated_changes["Debusine-Work-Request"],
            "https://example.com/debusine/Testing/work-request/100/",
        )

    def test_update_changes_file_without_workflow(self) -> None:
        directory = self.create_temporary_directory()
        changes_file = directory / "foo.changes"
        self.write_changes_file(changes_file, [])
        with changes_file.open("a") as f:
            f.write("\nDebusine-Workflow: foo\n")
        self.task._files = self.make_files_response(
            {"foo.changes": changes_file}
        )

        debusine_mock = self.mock_debusine(self.task)
        debusine_mock.base_api_url = "https://example.com/"
        debusine_mock.work_request_get.return_value = (
            create_work_request_response(id=100, workflow_root=None)
        )

        self.task.update_changes_file(directory)

        updated_changes = read_changes(changes_file)

        self.assertIn("Debusine-User", updated_changes)
        self.assertEqual(
            updated_changes["Debusine-User"], "https://example.com/-/users/123/"
        )
        self.assertNotIn("Debusine-Workflow", updated_changes)
        self.assertIn("Debusine-Work-Request", updated_changes)
        self.assertEqual(
            updated_changes["Debusine-Work-Request"],
            "https://example.com/debusine/Testing/work-request/100/",
        )

    def test_run_error(self) -> None:
        """run() raises an exception if signing fails."""
        key = Key.objects.create(
            purpose=Key.Purpose.OPENPGP_UPLOAD,
            fingerprint="0" * 64,
            private_key={},
            public_key=b"",
        )
        download_directory = self.create_temporary_directory()
        execute_directory = self.create_temporary_directory()
        (changes_file := download_directory / "foo.changes").touch()
        task = Debsign(task_data={"unsigned": 2, "key": key.fingerprint})
        self.task_set_current(task)
        task._files = self.make_files_response({"foo.changes": changes_file})
        task._key = key
        task._permission_check = AssetPermissionCheckResponse(
            has_permission=True, username="testuser", user_id=123, resource={}
        )
        debusine_mock = self.mock_debusine(task)
        debusine_mock.base_api_url = "https://example.com/"
        debusine_mock.work_request_get.return_value = (
            create_work_request_response()
        )
        task.prepare_to_run(download_directory, execute_directory)

        with (
            mock.patch(
                "debusine.signing.tasks.debsign.sign",
                side_effect=ValueError("Boom"),
            ),
            self.assertRaisesRegex(ValueError, "Boom"),
        ):
            task.run(execute_directory)

        self.assertIsNone(task._signed_changes_path)

    def test_run_success(self) -> None:
        """run() records a successful result if signing succeeds."""
        key = Key.objects.create(
            purpose=Key.Purpose.OPENPGP_UPLOAD,
            fingerprint="0" * 64,
            private_key={},
            public_key=b"",
        )
        download_directory = self.create_temporary_directory()
        execute_directory = self.create_temporary_directory()
        (changes_file := download_directory / "foo.changes").touch()
        task = Debsign(task_data={"unsigned": 2, "key": key.fingerprint})
        self.task_set_current(task)
        task._files = self.make_files_response({"foo.changes": changes_file})
        task._key = key
        task._permission_check = AssetPermissionCheckResponse(
            has_permission=True,
            username="testuser",
            user_id=123,
            resource={"package": "foo"},
        )
        debusine_mock = self.mock_debusine(task)
        debusine_mock.base_api_url = "https://example.com/"
        debusine_mock.work_request_get.return_value = (
            create_work_request_response()
        )
        task.prepare_to_run(download_directory, execute_directory)

        with mock.patch("debusine.signing.tasks.debsign.sign") as mock_sign:
            self.assertEqual(
                task.run(execute_directory), WorkRequestResults.SUCCESS
            )

        self.assertEqual(
            task._signed_changes_path,
            execute_directory / "output" / "foo.changes",
        )
        mock_sign.assert_called_once_with(
            [key],
            execute_directory / "input" / "foo.changes",
            execute_directory / "output" / "foo.changes",
            SigningMode.DEBSIGN,
            task.work_request_id,
            username="testuser",
            user_id=123,
            resource={"package": "foo"},
            log_file=mock.ANY,
        )


class DebsignIntegrationTests(
    SigningTaskTestCase[Debsign], TestCase, TransactionTestCase
):
    """Integration tests for the :py:class:`Debsign` task."""

    def generate_signing_key(self, service_private_key: PrivateKey) -> Key:
        """Generate a signing key."""
        with (
            override_settings(
                DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
            ),
            open(os.devnull, "wb") as log_file,
        ):
            key = Key.objects.generate(
                Key.Purpose.OPENPGP_UPLOAD, "An OpenPGP key", 1, log_file
            )
        return key

    def make_upload(self) -> tuple[Path, Path, Upload]:
        """Prepare a minimal upload; return (dsc, changes, artifact)."""
        temp_path = self.create_temporary_directory()
        dsc_path = temp_path / "foo.dsc"
        self.write_dsc_file(dsc_path, [])
        changes_path = temp_path / "foo.changes"
        self.write_changes_file(changes_path, [dsc_path])
        upload = Upload.create(changes_file=changes_path)
        return dsc_path, changes_path, upload

    def test_execute_success(self) -> None:
        """Integration test: sign using a real OpenPGP key."""
        service_private_key = PrivateKey.generate()
        key = self.generate_signing_key(service_private_key)
        dsc_path, changes_path, upload = self.make_upload()
        upload_id = 2
        upload_response = self.make_artifact_response(upload, upload_id)
        uploaded_paths: list[Path] = []
        uploaded_upload: Upload | None = None
        output_path = self.create_temporary_directory()
        uploaded_artifact_ids = {
            ArtifactCategory.UPLOAD: 3,
            ArtifactCategory.WORK_REQUEST_DEBUG_LOGS: 4,
        }

        def download_artifact(
            artifact_id: int, destination: Path, **kwargs: Any
        ) -> ArtifactResponse:
            assert artifact_id == upload_id
            shutil.copy(dsc_path, destination)
            shutil.copy(changes_path, destination)
            return upload_response

        def upload_artifact(
            local_artifact: LocalArtifact[Any], **kwargs: Any
        ) -> RemoteArtifact:
            nonlocal uploaded_upload
            for path in local_artifact.files.values():
                uploaded_paths.append(path)
                shutil.copy(path, output_path)
            if local_artifact.category == ArtifactCategory.UPLOAD:
                [changes_path] = [
                    path
                    for path in local_artifact.files.values()
                    if path.name.endswith(".changes")
                ]
                uploaded_upload = Upload.create(changes_file=changes_path)
            return create_remote_artifact(
                id=uploaded_artifact_ids[local_artifact.category],
                workspace="testing",
            )

        task = Debsign(
            task_data={"unsigned": upload_id, "key": key.fingerprint},
            dynamic_task_data={
                "unsigned_id": upload_id,
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.side_effect = download_artifact
        debusine_mock.upload_artifact.side_effect = upload_artifact
        debusine_mock.asset_permission_check.return_value = (
            AssetPermissionCheckResponse(
                has_permission=True,
                username="testuser",
                user_id=123,
                resource={},
            )
        )
        debusine_mock.base_api_url = "https://example.com/"
        debusine_mock.work_request_get.return_value = (
            create_work_request_response()
        )

        with override_settings(
            DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
        ):
            self.assertEqual(task.execute(), WorkRequestResults.SUCCESS)

        debusine_mock.asset_permission_check.assert_called_once_with(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=f"openpgp:{key.fingerprint}",
            permission_name="sign_with",
            artifact_id=upload_id,
            work_request_id=task.work_request_id,
            workspace=task.workspace_name,
        )

        # The task uploaded an artifact with the expected properties.
        self.assertEqual(len(uploaded_paths), 3)
        debusine_mock.upload_artifact.assert_has_calls(
            [
                mock.call(
                    uploaded_upload,
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
                mock.call(
                    WorkRequestDebugLogs(
                        category=WorkRequestDebugLogs._category,
                        data=EmptyArtifactData(),
                        files={
                            "cmd-output.log": (
                                task.log_directory / "cmd-output.log"
                            )
                        },
                        content_types={
                            "cmd-output.log": "text/plain; charset=utf-8"
                        },
                    ),
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
            ]
        )
        debusine_mock.relation_create.assert_any_call(
            uploaded_artifact_ids[ArtifactCategory.UPLOAD],
            upload_id,
            "relates-to",
        )

        # The file was properly signed.
        with (
            TemporaryDirectory(prefix="debusine-tests-") as tmp,
            gpg_ephemeral_context(Path(tmp)) as ctx,
        ):
            import_result = ctx.key_import(bytes(key.public_key))
            assert getattr(import_result, "imported", 0) == 1
            subprocess.run(
                [
                    "dscverify",
                    "--keyring",
                    Path(ctx.home_dir, "pubring.kbx"),
                    output_path / "foo.changes",
                ],
                env={**os.environ, "GNUPGHOME": ctx.home_dir},
                check=True,
                stdout=subprocess.DEVNULL,
            )

        changes_contents = (output_path / "foo.changes").read_text()
        self.assertIn("Debusine-Workflow", changes_contents)

    def test_execute_failure(self) -> None:
        """If execution fails, the task does not upload an output artifact."""
        service_private_key = PrivateKey.generate()
        key = self.generate_signing_key(service_private_key)
        dsc_path, changes_path, upload = self.make_upload()
        upload_id = 2
        upload_response = self.make_artifact_response(upload, upload_id)

        def download_artifact(
            artifact_id: int, destination: Path, **kwargs: Any
        ) -> ArtifactResponse:
            assert artifact_id == upload_id
            shutil.copy(dsc_path, destination)
            shutil.copy(changes_path, destination)
            return upload_response

        task = Debsign(
            task_data={
                "unsigned": upload_id,
                "key": key.fingerprint,
            },
            dynamic_task_data={
                "unsigned_id": upload_id,
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.side_effect = download_artifact
        debusine_mock.base_api_url = "https://example.com/"
        debusine_mock.work_request_get.return_value = (
            create_work_request_response()
        )

        with mock.patch(
            "debusine.signing.tasks.debsign.sign",
            side_effect=Exception("Signing failed"),
        ):
            self.assertEqual(task.execute(), WorkRequestResults.FAILURE)

        # Only a debug logs artifact was uploaded, not any signing output
        debusine_mock.upload_artifact.assert_called_once_with(
            WorkRequestDebugLogs(
                category=WorkRequestDebugLogs._category,
                data=EmptyArtifactData(),
                files={
                    "stages.log": (task.log_directory / "stages.log"),
                    "cmd-output.log": (task.log_directory / "cmd-output.log"),
                    "execution.log": (task.log_directory / "execution.log"),
                },
                content_types={
                    "stages.log": "text/plain; charset=utf-8",
                    "cmd-output.log": "text/plain; charset=utf-8",
                    "execution.log": "text/plain; charset=utf-8",
                },
            ),
            workspace=task.workspace_name,
            work_request=task.work_request_id,
        )
